# -*- coding: utf-8 -*-
"""Tools to manage the container structure"""

import os
import subprocess

from udocker import is_genstr
from udocker.config import Config
from udocker.msg import Msg
from udocker.utils.fileutil import FileUtil
from udocker.utils.uprocess import Uprocess
from udocker.helper.unique import Unique
from udocker.helper.hostinfo import HostInfo


class ContainerStructure(object):
    """Docker container structure.
    Creation of a container filesystem from a repository image.
    Creation of a container filesystem from a layer tar file.
    Access to container metadata.
    """

    def __init__(self, localrepo, container_id=None):
        self.localrepo = localrepo
        self.container_id = container_id
        self.tag = ""
        self.imagerepo = ""

    def get_container_attr(self):
        """Get container directory and JSON metadata by id or name"""
        if Config.conf['location']:
            container_dir = ""
            container_json = []
        else:
            container_dir = self.localrepo.cd_container(self.container_id)
            if not container_dir:
                Msg().err("Error: container id or name not found")
                return(False, False)

            fjson = container_dir + "/container.json"
            container_json = self.localrepo.load_json(fjson)
            if not container_json:
                Msg().err("Error: invalid container json metadata")
                return(False, False)

        return(container_dir, container_json)

    def get_container_meta(self, param, default, container_json):
        """Get the container metadata from the container"""
        confidx = ""
        if "config" in container_json:
            confidx = "config"
        elif "container_config" in container_json:
            confidx = "container_config"
        if container_json[confidx]  and param in container_json[confidx]:
            if container_json[confidx][param] is None:
                pass
            elif (is_genstr(container_json[confidx][param]) and
                  (isinstance(default, (list, tuple)))):
                return container_json[confidx][param].strip().split()
            elif (is_genstr(default) and (
                    isinstance(container_json[confidx][param], (list, tuple)))):
                return " ".join(container_json[confidx][param])
            elif (is_genstr(default) and (
                    isinstance(container_json[confidx][param], dict))):
                return self._dict_to_str(container_json[confidx][param])
            elif (isinstance(default, list) and (
                    isinstance(container_json[confidx][param], dict))):
                return self._dict_to_list(container_json[confidx][param])
            else:
                return container_json[confidx][param]
        return default

    # DEBUG
    def _dict_to_str(self, in_dict):
        """Convert dict to str"""
        out_str = ""
        for (key, val) in in_dict.items():
            out_str += f"{str(key)}:{str(val)} "
        return out_str

    def _dict_to_list(self, in_dict):
        """Convert dict to list"""
        out_list = []
        for (key, val) in in_dict.items():
            out_list.append(f"{str(key)}:{str(val)}")
        return out_list

    def _chk_container_root(self, container_id=None):
        """Check container ROOT sanity"""
        if container_id:
            container_dir = self.localrepo.cd_container(container_id)
        else:
            container_dir = self.localrepo.cd_container(self.container_id)
        if not container_dir:
            return 0
        container_root = container_dir + "/ROOT"
        check_list = ["/lib", "/bin", "/etc", "/tmp", "/var", "/usr", "/sys",
                      "/dev", "/data", "/home", "/system", "/root", "/proc", ]
        found = 0
        for f_path in check_list:
            if os.path.exists(container_root + f_path):
                found += 1
        return found

    def create_fromimage(self, imagerepo, tag):
        """Create a container from an image in the repository.
        Since images are stored as layers in tar files, this
        step consists in extracting those layers into a ROOT
        directory in the appropriate sequence.
        first argument: imagerepo
        second argument: image tag in that repo
        """
        self.imagerepo = imagerepo
        self.tag = tag
        image_dir = self.localrepo.cd_imagerepo(self.imagerepo, self.tag)
        if not image_dir:
            Msg().err("Error: create container: imagerepo is invalid")
            return False

        (container_json, layer_files) = self.localrepo.get_image_attributes()
        if not container_json:
            Msg().err("Error: create container: getting layers or json")
            return False

        if not self.container_id:
            self.container_id = Unique().uuid(os.path.basename(self.imagerepo))

        container_dir = self.localrepo.setup_container(
            self.imagerepo, self.tag, self.container_id)
        if not container_dir:
            Msg().err("Error: create container: setting up container")
            return False

        self.localrepo.save_json(
            container_dir + "/container.json", container_json)
        status = self._untar_layers(layer_files, container_dir + "/ROOT")
        if not status:
            Msg().err("Error: creating container:", self.container_id)
        elif not self._chk_container_root():
            Msg().out("Warning: check container content:", self.container_id,
                      l=Msg.WAR)

        return self.container_id

    def create_fromlayer(self, imagerepo, tag, layer_file, container_json):
        """Create a container from a layer file exported by Docker.
        """
        self.imagerepo = imagerepo
        self.tag = tag
        if not self.container_id:
            self.container_id = Unique().uuid(os.path.basename(self.imagerepo))

        if not container_json:
            Msg().err("Error: create container: getting json")
            return False

        container_dir = self.localrepo.setup_container(
            self.imagerepo, self.tag, self.container_id)
        if not container_dir:
            Msg().err("Error: create container: setting up")
            return False

        fjson = container_dir + "/container.json"
        self.localrepo.save_json(fjson, container_json)
        status = self._untar_layers([layer_file, ], container_dir + "/ROOT")
        if not status:
            Msg().err("Error: creating container:", self.container_id)
        elif not self._chk_container_root():
            Msg().out("Warning: check container content:", self.container_id,
                      l=Msg.WAR)

        return self.container_id

    def clone_fromfile(self, clone_file):
        """Create a cloned container from a file containing a cloned container
        exported by udocker.
        """
        if not self.container_id:
            self.container_id = Unique().uuid(os.path.basename(self.imagerepo))

        container_dir = self.localrepo.setup_container(
            "CLONING", "inprogress", self.container_id)
        if not container_dir:
            Msg().err("Error: create container: setting up")
            return False

        status = self._untar_layers([clone_file, ], container_dir)
        if not status:
            Msg().err("Error: creating container clone:", self.container_id)
        elif not self._chk_container_root():
            Msg().out("Warning: check container content:", self.container_id,
                      l=Msg.WAR)

        return self.container_id

    def _apply_whiteouts(self, tarf, destdir):
        """The layered filesystem of docker uses whiteout files
        to identify files or directories to be removed.
        The format is .wh.<filename>
        """
        verbose = ""
        if Msg.level >= Msg.VER:
            verbose = 'v'
            Msg().out("Info: applying whiteouts:", tarf, l=Msg.INF)
        wildcards = ["--wildcards", ]
        if not HostInfo().cmd_has_option("tar", wildcards[0]):
            wildcards = []
        cmd = ["tar", "t" + verbose] + wildcards + ["-f", tarf, r"*/.wh.*"]
        whiteouts = Uprocess().get_output(cmd, True)
        if not whiteouts:
            return
        for wh_filename in whiteouts.split('\n'):
            if wh_filename:
                wh_basename = os.path.basename(wh_filename.strip())
                wh_dirname = os.path.dirname(wh_filename)
                if wh_basename == ".wh..wh..opq":
                    if not os.path.isdir(destdir + '/' + wh_dirname):
                        continue
                    for f_name in os.listdir(destdir + '/' + wh_dirname):
                        rm_filename = destdir + '/' \
                            + wh_dirname + '/' + f_name
                        FileUtil(rm_filename).remove(recursive=True)
                elif wh_basename.startswith(".wh."):
                    rm_filename = destdir + '/' \
                        + wh_dirname + '/' \
                        + wh_basename.replace(".wh.", "", 1)
                    FileUtil(rm_filename).remove(recursive=True)
        return

    def _untar_layers(self, tarfiles, destdir):
        """Untar all container layers. Each layer is extracted
        and permissions are changed to avoid file permission
        issues when extracting the next layer.
        """
        if not (tarfiles and destdir):
            return False
        status = True
        gid = str(HostInfo.gid)
        optional_flags = ["--wildcards", "--delay-directory-restore", ]
        for option in optional_flags:
            if not HostInfo().cmd_has_option("tar", option):
                optional_flags.remove(option)
        for tarf in tarfiles:
            if tarf != '-':
                self._apply_whiteouts(tarf, destdir)
            verbose = ''
            if Msg.level >= Msg.VER:
                verbose = 'v'
                Msg().out("Info: extracting:", tarf, l=Msg.INF)
            cmd = ["tar", "-C", destdir, "-x" + verbose,
                   "--one-file-system", "--no-same-owner", "--overwrite",
                   "--exclude=dev/*", "--exclude=etc/udev/devices/*",
                   "--no-same-permissions", r"--exclude=.wh.*",
                   ] + optional_flags + ["-f", tarf]
            if subprocess.call(cmd, stderr=Msg.chlderr, close_fds=True):
                Msg().err("Error: while extracting image layer")
                status = False
            cmd = ["find", destdir,
                   "(", "-type", "d", "!", "-perm", "-u=x", "-exec",
                   "chmod", "u+x", "{}", ";", ")", ",",
                   "(", "!", "-perm", "-u=w", "-exec", "chmod",
                   "u+w", "{}", ";", ")", ",",
                   "(", "!", "-perm", "-u=r", "-exec", "chmod",
                   "u+r", "{}", ";", ")", ",",
                   "(", "!", "-gid", gid, "-exec", "chgrp", gid,
                   "{}", ";", ")", ",",
                   "(", "-name", ".wh.*", "-exec", "rm", "-f",
                   "--preserve-root", "{}", ";", ")"]
            if subprocess.call(cmd, stderr=Msg.chlderr, close_fds=True):
                status = False
                Msg().err("Error: while modifying attributes of image layer")
        return status

    def export_tofile(self, clone_file):
        """Export a container creating a tar file of the rootfs
        """
        cont_dir = self.localrepo.cd_container(self.container_id)
        if not cont_dir:
            Msg().err("Error: container not found:", self.container_id)
            return False

        status = FileUtil(cont_dir + "/ROOT").tar(clone_file)
        if not status:
            Msg().err("Error: exporting container file system:",
                      self.container_id)

        return self.container_id

    def clone_tofile(self, clone_file):
        """Create a cloned container tar file containing both the rootfs
        and all udocker control files. This is udocker specific.
        """
        container_dir = self.localrepo.cd_container(self.container_id)
        if not container_dir:
            Msg().err("Error: container not found:", self.container_id)
            return False

        status = FileUtil(container_dir).tar(clone_file)
        if not status:
            Msg().err("Error: export container as clone:", self.container_id)

        return self.container_id

    def clone(self):
        """Clone a container by creating a complete copy
        """
        source_container_dir = self.localrepo.cd_container(self.container_id)
        if not source_container_dir:
            Msg().err("Error: source container not found:", self.container_id)
            return False

        dest_container_id = Unique().uuid(os.path.basename(self.imagerepo))
        dest_container_dir = self.localrepo.setup_container(
            "CLONING", "inprogress", dest_container_id)
        if not dest_container_dir:
            Msg().err("Error: create destination container: setting up")
            return False

        status = FileUtil(source_container_dir).copydir(dest_container_dir)
        if not status:
            Msg().err("Error: creating container:", dest_container_id)
            return False

        if not self._chk_container_root(dest_container_id):
            Msg().out("Warning: check container content:", dest_container_id,
                      l=Msg.WAR)

        return dest_container_id



#!/usr/bin/env python
"""
udocker unit tests: CommonLocalFileApi
"""

from unittest import TestCase, main
from unittest.mock import Mock, patch
from udocker.commonlocalfile import CommonLocalFileApi


class CommonLocalFileApiTestCase(TestCase):
    """Test CommonLocalFileApi()."""

    def setUp(self):
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    def test_01_init(self):
        """Test01 CommonLocalFileApi() constructor"""
        clfapi = CommonLocalFileApi(self.local)
        self.assertEqual(clfapi.localrepo, self.local)

    @patch('udocker.commonlocalfile.FileUtil.copyto')
    @patch('udocker.commonlocalfile.os.rename')
    def test_02__move_layer_to_v1repo(self, mock_rename, mock_copy):
        """Test02 CommonLocalFileApi()._move_layer_to_v1repo()."""
        layer_id = "xxx"
        filepath = "yy"
        self.local.layersdir = "/home/.udocker"
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._move_layer_to_v1repo(filepath, layer_id)
        self.assertFalse(status)

        layer_id = "12345"
        filepath = "/home/.udocker/12345.json"
        self.local.layersdir = "/home/.udocker"
        mock_rename.return_value = None
        mock_copy.return_value = True
        self.local.add_image_layer.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._move_layer_to_v1repo(filepath, layer_id)
        self.assertTrue(status)
        self.assertTrue(mock_rename.called)
        self.assertTrue(self.local.add_image_layer.called)

        layer_id = "12345"
        filepath = "/home/.udocker/12345.layer.tar"
        self.local.layersdir = "/home/.udocker"
        mock_rename.return_value = None
        mock_copy.return_value = True
        self.local.add_image_layer.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._move_layer_to_v1repo(filepath, layer_id)
        self.assertTrue(status)
        self.assertTrue(mock_rename.called)
        self.assertTrue(self.local.add_image_layer.called)

    # def test_03__load_image_step2(self):
    #     """Test03 CommonLocalFileApi()._load_image_step2()."""

    @patch('udocker.commonlocalfile.Msg')
    def test_04__load_image(self, mock_msg):
        """Test04 CommonLocalFileApi()._load_image()."""
        mock_msg.level = 0
        structure = "12345"
        imagerepo = "/home/.udocker/images"
        tag = "v1"
        self.local.cd_imagerepo.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._load_image(structure, imagerepo, tag)
        self.assertEqual(status, [])

        self.local.cd_imagerepo.return_value = False
        self.local.setup_imagerepo.return_value = True
        self.local.setup_tag.return_value = False
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._load_image(structure, imagerepo, tag)
        self.assertEqual(status, [])

        self.local.cd_imagerepo.return_value = False
        self.local.setup_imagerepo.return_value = True
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = False
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._load_image(structure, imagerepo, tag)
        self.assertEqual(status, [])

        self.local.cd_imagerepo.return_value = False
        self.local.setup_imagerepo.return_value = True
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._load_image(structure, imagerepo, tag)
        self.assertEqual(status, None)

    @patch('udocker.commonlocalfile.Uprocess.call')
    def test_05__untar_saved_container(self, mock_ucall):
        """Test05 CommonLocalFileApi()._untar_saved_container()."""
        tarfile = "file.tar"
        destdir = "/home/.udocker/images"
        mock_ucall.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._untar_saved_container(tarfile, destdir)
        self.assertFalse(status)

        mock_ucall.return_value = False
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._untar_saved_container(tarfile, destdir)
        self.assertTrue(status)

    @patch('udocker.commonlocalfile.FileUtil.size')
    @patch('udocker.commonlocalfile.HostInfo.osversion')
    @patch('udocker.commonlocalfile.HostInfo.arch')
    def test_06_create_container_meta(self, mock_arch, mock_version,
                                      mock_size):
        """Test06 CommonLocalFileApi().create_container_meta()."""
        layer_id = "12345"
        comment = "created by my udocker"
        mock_arch.return_value = "x86_64"
        mock_version.return_value = "8"
        mock_size.return_value = 125
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.create_container_meta(layer_id, comment)
        self.assertEqual(status["id"], layer_id)
        self.assertEqual(status["comment"], comment)
        self.assertTrue(mock_arch.called)
        self.assertTrue(mock_version.called)
        self.assertTrue(mock_size.called)

    @patch('udocker.commonlocalfile.Msg')
    @patch('udocker.commonlocalfile.os.rename')
    @patch('udocker.commonlocalfile.FileUtil.copyto')
    @patch('udocker.commonlocalfile.Unique.layer_v1')
    @patch('udocker.commonlocalfile.os.path.exists')
    def test_07_import_toimage(self, mock_exists, mock_layerv1,
                               mock_copy, mock_rename, mock_msg):
        """Test07 CommonLocalFileApi().import_toimage()."""
        tarfile = "img.tar"
        imagerepo = "/home/.udocker/images"
        tag = "v1"
        move_tarball = True
        mock_msg.level = 0
        mock_exists.side_effect = [False, False]
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_toimage(tarfile, imagerepo, tag, move_tarball)
        self.assertFalse(status)

        mock_exists.side_effect = [True, False]
        self.local.setup_imagerepo.return_value = True
        self.local.cd_imagerepo.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_toimage(tarfile, imagerepo, tag, move_tarball)
        self.assertFalse(status)
        self.assertTrue(self.local.setup_imagerepo.called)
        self.assertTrue(self.local.cd_imagerepo.called)

        mock_exists.side_effect = [True, False]
        self.local.setup_imagerepo.return_value = True
        self.local.cd_imagerepo.return_value = False
        self.local.setup_tag.return_value = False
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_toimage(tarfile, imagerepo, tag, move_tarball)
        self.assertFalse(status)
        self.assertTrue(self.local.setup_imagerepo.called)
        self.assertTrue(self.local.cd_imagerepo.called)
        self.assertTrue(self.local.setup_tag.called)

        mock_exists.side_effect = [True, False]
        self.local.setup_imagerepo.return_value = True
        self.local.cd_imagerepo.return_value = False
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = False
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_toimage(tarfile, imagerepo, tag, move_tarball)
        self.assertFalse(status)
        self.assertTrue(self.local.setup_imagerepo.called)
        self.assertTrue(self.local.cd_imagerepo.called)
        self.assertTrue(self.local.setup_tag.called)

        tarfile = "img.tar"
        imagerepo = "/home/.udocker/images"
        tag = "v1"
        move_tarball = True
        mock_exists.side_effect = [True, False]
        self.local.setup_imagerepo.return_value = True
        self.local.cd_imagerepo.return_value = False
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = True
        mock_layerv1.return_value = "12345"
        mock_copy.return_value = False
        mock_rename.return_value = None
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_toimage(tarfile, imagerepo, tag, move_tarball)
        self.assertFalse(status)
        self.assertTrue(mock_layerv1.called)
        self.assertTrue(mock_rename.called)

        move_tarball = False
        mock_exists.side_effect = [True, True]
        self.local.setup_imagerepo.return_value = True
        self.local.cd_imagerepo.return_value = False
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = True
        self.local.add_image_layer.side_effect = [True, True]
        self.local.save_json.side_effect = [True, True]
        mock_layerv1.return_value = "12345"
        mock_copy.return_value = False
        mock_rename.return_value = None
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_toimage(tarfile, imagerepo, tag, move_tarball)
        self.assertEqual(status, "12345")
        self.assertTrue(mock_layerv1.called)

    @patch('udocker.commonlocalfile.Msg')
    @patch('udocker.commonlocalfile.ContainerStructure.create_fromlayer')
    @patch('udocker.commonlocalfile.Unique.layer_v1')
    @patch('udocker.commonlocalfile.os.path.exists')
    def test_08_import_tocontainer(self, mock_exists, mock_layerv1,
                                   mock_create, mock_msg):
        """Test08 CommonLocalFileApi().import_tocontainer()."""
        tarfile = ""
        imagerepo = ""
        tag = ""
        container_name = ""
        mock_msg.level = 0
        mock_exists.return_value = False
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_tocontainer(tarfile, imagerepo,
                                           tag, container_name)
        self.assertFalse(status)

        tarfile = "img.tar"
        imagerepo = "/home/.udocker/images"
        tag = "v1"
        container_name = "mycont"
        self.local.get_container_id.return_value = True
        mock_exists.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_tocontainer(tarfile, imagerepo,
                                           tag, container_name)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(mock_exists.called)
        self.assertFalse(status)

        tarfile = "img.tar"
        imagerepo = "/home/.udocker/images"
        tag = "v1"
        container_name = "mycont"
        self.local.get_container_id.return_value = False
        self.local.set_container_name.return_value = True
        mock_exists.return_value = True
        mock_layerv1.return_value = "12345"
        mock_create.return_value = "345"
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_tocontainer(tarfile, imagerepo,
                                           tag, container_name)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(mock_exists.called)
        self.assertTrue(mock_layerv1.called)
        self.assertTrue(mock_create.called)
        self.assertEqual(status, "345")

    @patch('udocker.commonlocalfile.Msg')
    @patch('udocker.commonlocalfile.ContainerStructure.clone_fromfile')
    @patch('udocker.commonlocalfile.os.path.exists')
    def test_09_import_clone(self, mock_exists, mock_clone, mock_msg):
        """Test09 CommonLocalFileApi().import_clone()."""
        tarfile = ""
        container_name = ""
        mock_msg.level = 0
        mock_exists.return_value = False
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_clone(tarfile, container_name)
        self.assertFalse(status)

        tarfile = "img.tar"
        container_name = "mycont"
        self.local.get_container_id.return_value = True
        mock_exists.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_clone(tarfile, container_name)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(mock_exists.called)
        self.assertFalse(status)

        tarfile = "img.tar"
        container_name = "mycont"
        self.local.get_container_id.return_value = False
        self.local.set_container_name.return_value = True
        mock_exists.return_value = True
        mock_clone.return_value = "345"
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.import_clone(tarfile, container_name)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(mock_exists.called)
        self.assertTrue(mock_clone.called)
        self.assertEqual(status, "345")

    @patch('udocker.commonlocalfile.Msg')
    @patch('udocker.commonlocalfile.ExecutionMode.set_mode')
    @patch('udocker.commonlocalfile.ExecutionMode.get_mode')
    @patch('udocker.commonlocalfile.ContainerStructure.clone')
    def test_10_clone_container(self, mock_clone,
                                mock_exget, mock_exset, mock_msg):
        """Test10 CommonLocalFileApi().clone_container()."""
        container_name = "mycont"
        container_id = ""
        mock_msg.level = 0
        self.local.get_container_id.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.clone_container(container_id, container_name)
        self.assertFalse(status)

        container_name = "mycont"
        container_id = ""
        self.local.get_container_id.return_value = False
        self.local.set_container_name.return_value = True
        mock_clone.return_value = "345"
        mock_exget.return_value = "P1"
        mock_exset.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.clone_container(container_id, container_name)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(self.local.set_container_name.called)
        self.assertTrue(mock_exget.called)
        self.assertFalse(mock_exset.called)
        self.assertTrue(mock_clone.called)
        self.assertEqual(status, "345")

        container_name = "mycont"
        container_id = ""
        self.local.get_container_id.return_value = False
        self.local.set_container_name.return_value = True
        mock_clone.return_value = "345"
        mock_exget.return_value = "F1"
        mock_exset.return_value = True
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi.clone_container(container_id, container_name)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(self.local.set_container_name.called)
        self.assertTrue(mock_exget.called)
        self.assertTrue(mock_exset.called)
        self.assertTrue(mock_clone.called)
        self.assertEqual(status, "345")

    @patch('udocker.commonlocalfile.os.path.exists')
    def test_11__get_imagedir_type(self, mock_exists):
        """Test11 CommonLocalFileApi()._get_imagedir_type()."""
        tmp_imagedir = "/home/.udocker/images/myimg"
        mock_exists.side_effect = [False, False]
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._get_imagedir_type(tmp_imagedir)
        self.assertEqual(status, "")

        mock_exists.side_effect = [True, False]
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._get_imagedir_type(tmp_imagedir)
        self.assertEqual(status, "OCI")

        mock_exists.side_effect = [False, True]
        clfapi = CommonLocalFileApi(self.local)
        status = clfapi._get_imagedir_type(tmp_imagedir)
        self.assertEqual(status, "Docker")

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: Uvolume
"""

from unittest import TestCase, main
from udocker.utils.uvolume import Uvolume


class UvolumeTestCase(TestCase):
    """Test Uvolume()."""

    def test_01_init(self):
        """Test01 Uvolume() constructor"""
        volm = '/data'
        uvol = Uvolume(volm)
        self.assertEqual(uvol.volume, volm)

    def test_02_cleanpath(self):
        """Test02 Uvolume().cleanpath()."""
        path = '/bin/ls'
        volm = '/data'
        uvol = Uvolume(volm)
        result = uvol.cleanpath(path)
        self.assertEqual(result, path)

        path2 = '//bin//ls'
        volm = '/data'
        uvol = Uvolume(volm)
        result = uvol.cleanpath(path2)
        self.assertEqual(result, path)

    def test_03_split(self):
        """Test03 Uvolume().split()."""
        volm = '/data:/contdata'
        uvol = Uvolume(volm)
        result = uvol.split()
        self.assertEqual(result, ('/data', '/contdata'))

        volm = '/data'
        uvol = Uvolume(volm)
        result = uvol.split()
        self.assertEqual(result, ('/data', '/data'))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: DockerIoAPI
"""

from unittest import TestCase, main
from unittest.mock import patch, Mock
from io import BytesIO as strio
from udocker.docker import DockerIoAPI
from udocker.config import Config


class DockerIoAPITestCase(TestCase):
    """Test DockerIoAPITest().
    Class to encapsulate the access to the Docker Hub service.
    Allows to search and download images from Docker Hub.
    """

    def setUp(self):
        Config().getconf()
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    @patch('udocker.docker.GetURL')
    def test_01_init(self, mock_geturl):
        """Test01 DockerIoAPI() constructor"""
        mock_geturl.return_value = None
        doia = DockerIoAPI(self.local)
        self.assertEqual(doia.index_url,
                         Config.conf['dockerio_index_url'])
        self.assertEqual(doia.registry_url,
                         Config.conf['dockerio_registry_url'])
        self.assertEqual(doia.v1_auth_header, "")
        self.assertEqual(doia.v2_auth_header, "")
        self.assertEqual(doia.v2_auth_token, "")
        self.assertEqual(doia.localrepo, self.local)
        self.assertTrue(doia.search_pause)
        self.assertEqual(doia.search_page, 0)
        self.assertFalse(doia.search_ended)
        self.assertTrue(mock_geturl.called)

    @patch('udocker.docker.GetURL')
    def test_02_set_proxy(self, mock_geturl):
        """Test02 DockerIoAPI().set_proxy()."""
        url = "socks5://user:pass@host:port"
        doia = DockerIoAPI(self.local)
        doia.set_proxy(url)
        self.assertTrue(mock_geturl.return_value.set_proxy.called_with(url))

    def test_03_set_registry(self):
        """Test03 DockerIoAPI().set_registry()."""
        doia = DockerIoAPI(self.local)
        doia.set_registry("https://registry-1.docker.io")
        self.assertEqual(doia.registry_url, "https://registry-1.docker.io")

    def test_04_set_index(self):
        """Test04 DockerIoAPI().set_index()."""
        doia = DockerIoAPI(self.local)
        doia.set_index("https://index.docker.io/v1")
        self.assertEqual(doia.index_url, "https://index.docker.io/v1")

    def test_05_is_repo_name(self):
        """Test05 DockerIoAPI().is_repo_name()."""
        doia = DockerIoAPI(self.local)
        self.assertFalse(doia.is_repo_name(""))
        self.assertFalse(doia.is_repo_name("socks5://user:pass@host:port"))
        self.assertFalse(doia.is_repo_name("/:"))
        self.assertTrue(doia.is_repo_name("1233/fasdfasdf:sdfasfd"))
        self.assertTrue(doia.is_repo_name("os-cli-centos7"))
        self.assertTrue(doia.is_repo_name("os-cli-centos7:latest"))
        self.assertTrue(doia.is_repo_name("lipcomputing/os-cli-centos7"))
        self.assertTrue(doia.is_repo_name("lipcomputing/os-cli-centos7:latest"))

    @patch('udocker.docker.GetURL.get_status_code')
    @patch('udocker.docker.GetURL.get')
    def test_06__get_url(self, mock_get, mock_getstatus):
        """Test06 DockerIoAPI()._get_url()."""
        args = ["http://some1.org"]
        kwargs = list()
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase"}
        buff = strio()
        mock_get.return_value = (hdr, buff)
        mock_getstatus.return_value = 200
        doia = DockerIoAPI(self.local)
        status = doia._get_url(args, kwargs)
        self.assertEqual(status, (hdr, buff))

        args = ["http://some1.org"]
        kwargs = list()
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 400 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0,
                    "location": "http://some1.org"}
        buff = strio()
        mock_get.return_value = (hdr, buff)
        mock_getstatus.return_value = 400
        doia = DockerIoAPI(self.local)
        status = doia._get_url(args, kwargs)
        self.assertEqual(status, (hdr, buff))

        args = ["http://some1.org"]
        kwargs = {"RETRY": True, "FOLLOW": True}
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 400 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_get.return_value = (hdr, buff)
        mock_getstatus.return_value = 400
        doia = DockerIoAPI(self.local)
        status = doia._get_url(args, kwargs)
        self.assertEqual(status, (hdr, buff))

    @patch.object(DockerIoAPI, '_get_url')
    @patch('udocker.docker.GetURL.get_status_code')
    @patch('udocker.docker.FileUtil.size')
    @patch('udocker.docker.GetURL.get_content_length')
    @patch('udocker.docker.ChkSUM.hash')
    def test_07__get_file(self, mock_hash, mock_getlength,
                          mock_fusize, mock_status, mock_geturl):
        """Test07 DockerIoAPI()._get_file()."""
        cks = "af98ca7807fd3859c5bd876004fa7e960cecebddb342de1bc7f3b0e6f7dab415"
        url = "http://some1.org/file1"
        fname = "/sha256:" + cks
        cache = 0
        mock_hash.return_value = cks
        doia = DockerIoAPI(self.local)
        status = doia._get_file(url, fname, cache)
        self.assertTrue(status)

        cks = "af98ca7807fd3859c5bd876004fa7e960cecebddb342de1bc7f3b0e6f7dab415"
        url = "http://some1.org/file1"
        fname = ""
        cache = 1
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 400 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_hash.return_value = cks
        mock_geturl.return_value = (hdr, buff)
        mock_getlength.return_value = 123
        mock_fusize.return_value = 123
        doia = DockerIoAPI(self.local)
        doia.curl.cache_support = True
        status = doia._get_file(url, fname, cache)
        self.assertTrue(status)

        cks = "af98ca7807fd3859c5bd876004fa7e960cecebddb342de1bc7f3b0e6f7dab415"
        url = "http://some1.org/file1"
        fname = cks + ".layer"
        cache = 0
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 400 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_hash.return_value = cks
        mock_geturl.return_value = (hdr, buff)
        mock_getlength.return_value = 123
        mock_fusize.return_value = 123
        mock_status.return_value = 200
        doia = DockerIoAPI(self.local)
        doia.curl.cache_support = False
        status = doia._get_file(url, fname, cache)
        self.assertTrue(status)

    def test_08__split_fields(self):
        """Test08 DockerIoAPI()._split_fields()."""
        buff = 'k1="v1",k2="v2"'
        doia = DockerIoAPI(self.local)
        status = doia._split_fields(buff)
        self.assertEqual(status, {"k1": "v1", "k2": "v2"})

    @patch.object(DockerIoAPI, '_get_url')
    def test_09_is_v1(self, mock_geturl):
        """Test09 DockerIoAPI().is_v1()."""
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_geturl.return_value = (hdr, buff)
        doia = DockerIoAPI(self.local)
        status = doia.is_v1()
        self.assertTrue(status)

        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 400 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_geturl.return_value = (hdr, buff)
        doia = DockerIoAPI(self.local)
        status = doia.is_v1()
        self.assertFalse(status)

    @patch.object(DockerIoAPI, '_get_url')
    def test_10_has_search_v1(self, mock_geturl):
        """Test10 DockerIoAPI().has_search_v1()."""
        url = "http://some1.org/file1"
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_geturl.return_value = (hdr, buff)
        doia = DockerIoAPI(self.local)
        status = doia.has_search_v1(url)
        self.assertTrue(status)

        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 400 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_geturl.return_value = (hdr, buff)
        doia = DockerIoAPI(self.local)
        status = doia.has_search_v1(url)
        self.assertFalse(status)

    @patch('udocker.docker.json.loads')
    @patch.object(DockerIoAPI, '_get_url')
    def test_11_get_v1_repo(self, mock_geturl, mock_jload):
        """Test11 DockerIoAPI().get_v1_repo"""
        imagerepo = "REPO"
        hdr = type('test', (object,), {})()
        hdr.data = {"x-docker-token": "12345"}
        buff = strio()
        mock_geturl.return_value = (hdr, buff)
        mock_jload.return_value = {"k1": "v1"}
        doia = DockerIoAPI(self.local)
        doia.index_url = "docker.io"
        status = doia.get_v1_repo(imagerepo)
        self.assertEqual(status, ({"x-docker-token": "12345"}, {"k1": "v1"}))

    def test_12__get_v1_auth(self):
        """Test12 DockerIoAPI()._get_v1_auth"""
        doia = DockerIoAPI(self.local)
        doia.v1_auth_header = "Not Empty"
        www_authenticate = ['Other Stuff']
        out = doia._get_v1_auth(www_authenticate)
        self.assertEqual(out, "")

        www_authenticate = ['Token']
        doia = DockerIoAPI(self.local)
        doia.v1_auth_header = "Not Empty"
        out = doia._get_v1_auth(www_authenticate)
        self.assertEqual(out, "Not Empty")

    @patch('udocker.docker.Msg')
    @patch('udocker.utils.curl.CurlHeader')
    @patch.object(DockerIoAPI, '_get_url')
    def test_13_get_v1_image_tags(self, mock_dgu, mock_hdr, mock_msg):
        """Test13 DockerIoAPI().get_v1_image_tags"""
        mock_msg.level = 0
        mock_dgu.return_value = (mock_hdr, [])
        endpoint = "docker.io"
        imagerepo = "REPO"
        doia = DockerIoAPI(self.local)
        out = doia.get_v1_image_tags(endpoint, imagerepo)
        self.assertIsInstance(out, list)

    @patch('udocker.docker.Msg')
    @patch('udocker.utils.curl.CurlHeader')
    @patch.object(DockerIoAPI, '_get_url')
    def test_14_get_v1_image_tag(self, mock_dgu, mock_hdr, mock_msg):
        """Test14 DockerIoAPI().get_v1_image_tag"""
        mock_msg.level = 0
        mock_dgu.return_value = (mock_hdr, [])
        endpoint = "docker.io"
        imagerepo = "REPO"
        tag = "TAG"
        doia = DockerIoAPI(self.local)
        out = doia.get_v1_image_tag(endpoint, imagerepo, tag)
        self.assertIsInstance(out, tuple)

    @patch('udocker.docker.Msg')
    @patch('udocker.utils.curl.CurlHeader')
    @patch.object(DockerIoAPI, '_get_url')
    def test_15_get_v1_image_ancestry(self, mock_dgu, mock_hdr, mock_msg):
        """Test15 DockerIoAPI().get_v1_image_ancestry"""
        mock_msg.level = 0
        mock_dgu.return_value = (mock_hdr, [])
        endpoint = "docker.io"
        image_id = "IMAGEID"
        doia = DockerIoAPI(self.local)
        out = doia.get_v1_image_ancestry(endpoint, image_id)
        self.assertIsInstance(out, tuple)

    @patch('udocker.docker.Msg')
    @patch.object(DockerIoAPI, '_get_file')
    def test_16_get_v1_image_json(self, mock_dgf, mock_msg):
        """Test16 DockerIoAPI().get_v1_image_json"""
        mock_msg.level = 0
        mock_dgf.return_value = True
        endpoint = "docker.io"
        layer_id = "LAYERID"
        doia = DockerIoAPI(self.local)
        status = doia.get_v1_image_json(endpoint, layer_id)
        self.assertTrue(status)

        mock_dgf.return_value = False
        doia = DockerIoAPI(self.local)
        status = doia.get_v1_image_json(endpoint, layer_id)
        self.assertFalse(status)

    @patch('udocker.docker.Msg')
    @patch.object(DockerIoAPI, '_get_file')
    def test_17_get_v1_image_layer(self, mock_dgf, mock_msg):
        """Test17 DockerIoAPI().get_v1_image_layer"""
        mock_msg.level = 0
        mock_dgf.return_value = True
        endpoint = "docker.io"
        layer_id = "LAYERID"
        doia = DockerIoAPI(self.local)
        status = doia.get_v1_image_layer(endpoint, layer_id)
        self.assertTrue(status)

        mock_dgf.return_value = False
        doia = DockerIoAPI(self.local)
        status = doia.get_v1_image_layer(endpoint, layer_id)
        self.assertFalse(status)

    @patch('udocker.docker.Msg')
    @patch.object(DockerIoAPI, '_get_file')
    def test_18_get_v1_layers_all(self, mock_dgf, mock_msg):
        """Test18 DockerIoAPI().get_v1_layers_all"""
        mock_msg.level = 0
        mock_dgf.return_value = True
        endpoint = "docker.io"
        layer_list = []
        doia = DockerIoAPI(self.local)
        out = doia.get_v1_layers_all(endpoint, layer_list)
        self.assertEqual(out, [])

        layer_list = ["a", "b"]
        doia = DockerIoAPI(self.local)
        out = doia.get_v1_layers_all(endpoint, layer_list)
        self.assertEqual(out, ['b.json', 'b.layer', 'a.json', 'a.layer'])

    @patch.object(DockerIoAPI, '_get_url')
    @patch('udocker.utils.curl.CurlHeader')
    @patch('udocker.docker.json.loads')
    def test_19__get_v2_auth(self, mock_jloads, mock_hdr, mock_dgu):
        """Test19 DockerIoAPI()._get_v2_auth"""
        fakedata = strio('token'.encode('utf-8'))
        www_authenticate = "Other Stuff"
        mock_dgu.return_value = (mock_hdr, fakedata)
        mock_jloads.return_value = {'token': 'YYY'}
        doia = DockerIoAPI(self.local)
        doia.v2_auth_header = "v2 Auth Header"
        doia.v2_auth_token = "v2 Auth Token"
        out = doia._get_v2_auth(www_authenticate, False)
        self.assertEqual(out, "")

        www_authenticate = "Bearer realm=REALM"
        mock_dgu.return_value = (mock_hdr, fakedata)
        mock_jloads.return_value = {'token': 'YYY'}
        doia = DockerIoAPI(self.local)
        out = doia._get_v2_auth(www_authenticate, False)
        self.assertEqual(out, "Authorization: Bearer YYY")

        www_authenticate = "BASIC realm=Sonatype Nexus Repository"
        mock_dgu.return_value = (mock_hdr, fakedata)
        mock_jloads.return_value = {'token': 'YYY'}
        doia = DockerIoAPI(self.local)
        out = doia._get_v2_auth(www_authenticate, False)
        self.assertEqual(out, "Authorization: Basic %s" % doia.v2_auth_token)

    def test_20_get_v2_login_token(self):
        """Test20 DockerIoAPI().get_v2_login_token"""
        doia = DockerIoAPI(self.local)
        out = doia.get_v2_login_token("username", "password")
        self.assertIsInstance(out, str)

        doia = DockerIoAPI(self.local)
        out = doia.get_v2_login_token("", "")
        self.assertEqual(out, "")

    def test_21_set_v2_login_token(self):
        """Test21 DockerIoAPI().set_v2_login_token"""
        doia = DockerIoAPI(self.local)
        doia.set_v2_login_token("BIG-FAT-TOKEN")
        self.assertEqual(doia.v2_auth_token, "BIG-FAT-TOKEN")

    @patch('udocker.utils.curl.CurlHeader')
    @patch.object(DockerIoAPI, '_get_url')
    def test_22_is_v2(self, mock_dgu, mock_hdr):
        """Test22 DockerIoAPI().is_v2"""
        mock_dgu.return_value = (mock_hdr, [])
        doia = DockerIoAPI(self.local)
        doia.registry_url = "http://www.docker.io"
        out = doia.is_v2()
        self.assertFalse(out)

        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_dgu.return_value = (hdr, buff)
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.is_v2()
        self.assertTrue(out)

    @patch.object(DockerIoAPI, '_get_url')
    def test_23_has_search_v2(self, mock_dgu):
        """Test23 DockerIoAPI().has_search_v2"""
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 400 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_dgu.return_value = (hdr, buff)
        doia = DockerIoAPI(self.local)
        doia.registry_url = "http://www.docker.io"
        out = doia.has_search_v2()
        self.assertFalse(out)

        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_dgu.return_value = (hdr, buff)
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.has_search_v2()
        self.assertTrue(out)

    @patch('udocker.docker.json.loads')
    @patch.object(DockerIoAPI, '_get_url')
    def test_24_get_v2_image_tags(self, mock_dgu, mock_jload):
        """Test24 DockerIoAPI().get_v2_image_tags"""
        imgrepo = "img1"
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_dgu.return_value = (hdr, buff)
        mock_jload.return_value = list()
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.get_v2_image_tags(imgrepo)
        self.assertEqual(out, list())

        imgrepo = "img1"
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        buff = strio()
        mock_dgu.return_value = (hdr, buff)
        mock_jload.return_value = ["tag1", "tag2"]
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.get_v2_image_tags(imgrepo)
        self.assertEqual(out, ["tag1", "tag2"])

    @patch('udocker.docker.Msg')
    @patch('udocker.utils.curl.CurlHeader')
    @patch.object(DockerIoAPI, '_get_url')
    def test_25_get_v2_image_manifest(self, mock_dgu, mock_hdr, mock_msg):
        """Test25 DockerIoAPI().get_v2_image_manifest"""
        mock_msg.level = 0
        imagerepo = "REPO"
        tag = "TAG"
        mock_dgu.return_value = (mock_hdr, [])
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.get_v2_image_manifest(imagerepo, tag)
        self.assertIsInstance(out, tuple)

    @patch('udocker.docker.Msg')
    @patch.object(DockerIoAPI, '_get_file')
    def test_26_get_v2_image_layer(self, mock_dgf, mock_msg):
        """Test26 DockerIoAPI().get_v2_image_layer"""
        mock_msg.level = 0
        imagerepo = "REPO"
        layer_id = "LAYERID"
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"

        mock_dgf.return_value = True
        doia = DockerIoAPI(self.local)
        out = doia.get_v2_image_layer(imagerepo, layer_id)
        self.assertTrue(out)

        mock_dgf.return_value = False
        doia = DockerIoAPI(self.local)
        out = doia.get_v2_image_layer(imagerepo, layer_id)
        self.assertFalse(out)

    @patch('udocker.docker.Msg')
    @patch.object(DockerIoAPI, 'get_v2_image_layer')
    def test_27_get_v2_layers_all(self, mock_v2il, mock_msg):
        """Test27 DockerIoAPI().get_v2_layers_all"""
        mock_msg.level = 0
        mock_v2il.return_value = False
        imagerepo = "REPO"
        fslayers = []
        doia = DockerIoAPI(self.local)
        out = doia.get_v2_layers_all(imagerepo, fslayers)
        self.assertEqual(out, [])

        mock_msg.level = 0
        mock_v2il.return_value = True
        imagerepo = "REPO"
        fslayers = [{"blobSum": "foolayername"}]
        doia = DockerIoAPI(self.local)
        out = doia.get_v2_layers_all(imagerepo, fslayers)
        self.assertEqual(out, ['foolayername'])

    @patch('udocker.docker.Msg')
    @patch('udocker.docker.GetURL.get_status_code')
    @patch.object(DockerIoAPI, 'get_v2_image_manifest')
    @patch.object(DockerIoAPI, 'get_v2_layers_all')
    @patch.object(DockerIoAPI, '_get_url')
    def test_28_get_v2(self, mock_dgu, mock_dgv2, mock_manif,
                       mock_getstatus, mock_msg):
        """Test28 DockerIoAPI().get_v2"""
        imgrepo = "img1"
        hdr = type('test', (object,), {})()
        hdr_data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0}
        hdr.data = hdr_data
        buff = strio()
        manifest = list()
        mock_msg.level = 0
        mock_dgu.return_value = (hdr, buff)
        mock_manif.return_value = (hdr_data, manifest)
        mock_dgv2.return_value = []
        mock_getstatus.return_value = 400
        tag = "TAG"
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.get_v2(imgrepo, tag)
        self.assertEqual(out, [])

        mock_dgu.return_value = (hdr, buff)
        mock_manif.return_value = (hdr_data, manifest)
        mock_dgv2.return_value = []
        mock_getstatus.return_value = 200
        self.local.setup_tag.return_value = False
        self.local.set_version.return_value = True
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.get_v2(imgrepo, tag)
        self.assertEqual(out, [])

        manifest = {
            "fsLayers": ({"blobSum": "foolayername"},),
            "history": ({"v1Compatibility": '["foojsonstring"]'},)
        }
        mock_dgu.return_value = (hdr, buff)
        mock_manif.return_value = (hdr_data, manifest)
        mock_dgv2.return_value = ["foolayername"]
        mock_getstatus.return_value = 200
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = True
        self.local.save_json.return_value = None
        doia = DockerIoAPI(self.local)
        doia.registry_url = "https://registry-1.docker.io"
        out = doia.get_v2(imgrepo, tag)
        self.assertEqual(out, ["foolayername"])

    def test_29__get_v1_id_from_tags(self):
        """Test29 DockerIoAPI()._get_v1_id_from_tags"""
        tobj = {"tag1": "t1"}
        tag = "tag1"
        doia = DockerIoAPI(self.local)
        out = doia._get_v1_id_from_tags(tobj, tag)
        self.assertEqual(out, "t1")

        tobj = [{"name": "tag1", "layer": "l1"}]
        tag = "tag1"
        doia = DockerIoAPI(self.local)
        out = doia._get_v1_id_from_tags(tobj, tag)
        self.assertEqual(out, "l1")

    def test_30__get_v1_id_from_images(self):
        """Test30 DockerIoAPI()._get_v1_id_from_images"""
        imgarr = list()
        shortid = ""
        doia = DockerIoAPI(self.local)
        out = doia._get_v1_id_from_images(imgarr, shortid)
        self.assertEqual(out, "")

        imgarr = [{"id": "1234567890"}]
        shortid = "12345678"
        doia = DockerIoAPI(self.local)
        out = doia._get_v1_id_from_images(imgarr, shortid)
        self.assertEqual(out, "1234567890")

    @patch('udocker.docker.Msg')
    @patch('udocker.docker.GetURL.get_status_code')
    @patch.object(DockerIoAPI, 'get_v1_layers_all')
    @patch.object(DockerIoAPI, 'get_v1_image_ancestry')
    @patch.object(DockerIoAPI, '_get_v1_id_from_images')
    @patch.object(DockerIoAPI, '_get_v1_id_from_tags')
    @patch.object(DockerIoAPI, 'get_v1_image_tags')
    @patch.object(DockerIoAPI, 'get_v1_repo')
    def test_31_get_v1(self, mock_dgv1repo, mock_v1imgtag,
                       mock_v1idtag, mock_v1idimg, mock_v1ances,
                       mock_v1layer, mock_status, mock_msg):
        """Test31 DockerIoAPI().get_v1"""
        imgarr = [{"id": "1234567890"}]
        imagerepo = "REPO"
        tag = "TAG"
        hdr = type('test', (object,), {})()
        hdr_data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0,
                    "x-docker-endpoints": "https://registry-1.docker.io"}
        hdr.data = hdr_data
        mock_msg.level = 0
        mock_dgv1repo.return_value = (hdr_data, imgarr)
        mock_status.return_value = 401
        doia = DockerIoAPI(self.local)
        out = doia.get_v1(imagerepo, tag)
        self.assertEqual(out, [])

        mock_v1imgtag.return_value = [{"name": "tag1"}]
        mock_v1idtag.return_value = ""
        mock_dgv1repo.return_value = (hdr_data, imgarr)
        mock_status.return_value = 200
        doia = DockerIoAPI(self.local)
        out = doia.get_v1(imagerepo, tag)
        self.assertEqual(out, [])

        mock_v1imgtag.return_value = [{"name": "tag1"}]
        mock_v1idtag.return_value = "123456"
        mock_v1idimg.return_value = ""
        mock_dgv1repo.return_value = (hdr_data, imgarr)
        mock_status.return_value = 200
        doia = DockerIoAPI(self.local)
        out = doia.get_v1(imagerepo, tag)
        self.assertEqual(out, [])

        mock_v1imgtag.return_value = [{"name": "tag1"}]
        mock_v1idtag.return_value = "123456"
        mock_v1idimg.return_value = "123456789"
        mock_dgv1repo.return_value = (hdr_data, imgarr)
        mock_status.return_value = 200
        self.local.setup_tag.return_value = False
        self.local.set_version.return_value = True
        doia = DockerIoAPI(self.local)
        out = doia.get_v1(imagerepo, tag)
        self.assertEqual(out, [])

        mock_v1imgtag.return_value = [{"name": "tag1"}]
        mock_v1idtag.return_value = "123456"
        mock_v1idimg.return_value = "123456789"
        mock_dgv1repo.return_value = (hdr_data, imgarr)
        mock_status.return_value = 200
        mock_v1ances.return_value = ("x", "")
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = True
        self.local.save_json.return_value = None
        doia = DockerIoAPI(self.local)
        out = doia.get_v1(imagerepo, tag)
        self.assertEqual(out, [])

        mock_v1imgtag.return_value = [{"name": "tag1"}]
        mock_v1idtag.return_value = "123456"
        mock_v1idimg.return_value = "123456789"
        mock_dgv1repo.return_value = (hdr_data, imgarr)
        mock_status.return_value = 200
        mock_v1ances.return_value = ("x", "123")
        mock_v1layer.return_value = ["file1"]
        self.local.setup_tag.return_value = True
        self.local.set_version.return_value = True
        self.local.save_json.return_value = None
        doia = DockerIoAPI(self.local)
        out = doia.get_v1(imagerepo, tag)
        self.assertEqual(out, ["file1"])

    def test_32__parse_imagerepo(self):
        """Test32 DockerIoAPI()._parse_imagerepo"""
        imagerepo = "https://hub.docker.com/_/mysql"
        doia = DockerIoAPI(self.local)
        out = doia._parse_imagerepo(imagerepo)
        self.assertEqual(out, (imagerepo, "https://hub.docker.com/_/mysql"))

    @patch.object(DockerIoAPI, 'get_v1')
    @patch.object(DockerIoAPI, 'get_v2')
    @patch.object(DockerIoAPI, 'is_v2')
    @patch.object(DockerIoAPI, '_parse_imagerepo')
    def test_33_get(self, mock_parse, mock_isv2, mock_getv2, mock_getv1):
        """Test33 DockerIoAPI().get"""
        imagerepo = "REPO"
        tag = "TAG"
        mock_parse.return_value = ("REPO", "https://registry-1.docker.io")
        self.local.cd_imagerepo.return_value = False
        self.local.setup_imagerepo.return_value = True
        self.local.del_imagerepo.return_value = False
        mock_isv2.return_value = False
        mock_getv1.return_value = ["a", "b"]
        doia = DockerIoAPI(self.local)
        out = doia.get(imagerepo, tag)
        self.assertEqual(out, ["a", "b"])

        mock_parse.return_value = ("REPO", "https://registry-1.docker.io")
        self.local.cd_imagerepo.return_value = True
        self.local.setup_imagerepo.return_value = True
        self.local.del_imagerepo.return_value = False
        mock_isv2.return_value = True
        mock_getv2.return_value = ["a", "b"]
        doia = DockerIoAPI(self.local)
        out = doia.get(imagerepo, tag)
        self.assertEqual(out, ["a", "b"])

    @patch.object(DockerIoAPI, 'get_v1_image_tags')
    @patch.object(DockerIoAPI, 'get_v2_image_tags')
    @patch.object(DockerIoAPI, 'is_v2')
    def test_34_get_tags(self, mock_isv2, mock_getv2, mock_getv1):
        """Test34 DockerIoAPI().get_tags"""
        imagerepo = "REPO"
        mock_isv2.return_value = False
        mock_getv1.return_value = ["a", "b"]
        doia = DockerIoAPI(self.local)
        out = doia.get_tags(imagerepo)
        self.assertEqual(out, ["a", "b"])

        mock_isv2.return_value = True
        mock_getv2.return_value = ["a", "b"]
        doia = DockerIoAPI(self.local)
        out = doia.get_tags(imagerepo)
        self.assertEqual(out, ["a", "b"])

    def test_35_search_init(self):
        """Test35 DockerIoAPI().search_init"""
        doia = DockerIoAPI(self.local)
        doia.search_init("PAUSE")
        self.assertEqual(doia.search_pause, "PAUSE")
        self.assertEqual(doia.search_page, 0)
        self.assertEqual(doia.search_ended, False)

    @patch.object(DockerIoAPI, '_get_url')
    @patch('udocker.docker.json.loads')
    def test_36_search_get_page_v1(self, mock_jload, mock_dgu):
        """Test36 DockerIoAPI().set_index"""
        hdr = type('test', (object,), {})()
        hdr_data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0,
                    "x-docker-endpoints": "https://registry-1.docker.io"}
        hdr.data = hdr_data
        buff = strio()
        mock_dgu.return_value = (hdr, buff)
        mock_jload.return_value = {"page": 1, "num_pages": 1}
        url = "https://registry-1.docker.io"
        doia = DockerIoAPI(self.local)
        out = doia.search_get_page_v1("SOMETHING", url)
        self.assertEqual(out, {"page": 1, "num_pages": 1})

    @patch.object(DockerIoAPI, '_get_url')
    @patch('udocker.docker.json.loads')
    def test_37_search_get_page_v2(self, mock_jload, mock_dgu):
        """Test37 DockerIoAPI().search_get_page_v2"""
        hdr = type('test', (object,), {})()
        hdr_data = {"content-length": 10,
                    "X-ND-HTTPSTATUS": "HTTP-Version 200 Reason-Phrase",
                    "X-ND-CURLSTATUS": 0,
                    "x-docker-endpoints": "https://registry-1.docker.io"}
        hdr.data = hdr_data
        buff = strio()
        mock_dgu.return_value = (hdr, buff)
        mock_jload.return_value = {"count": 1, "num_pages": 1}
        url = "https://registry-1.docker.io"
        doia = DockerIoAPI(self.local)
        out = doia.search_get_page_v2("SOMETHING", url)
        self.assertEqual(out, {"count": 1, "num_pages": 1})

    @patch.object(DockerIoAPI, 'search_get_page_v1')
    @patch.object(DockerIoAPI, 'search_get_page_v2')
    @patch.object(DockerIoAPI, 'has_search_v1')
    @patch.object(DockerIoAPI, 'has_search_v2')
    def test_38_search_get_page(self, mock_searchv2, mock_searchv1,
                                mock_getv2, mock_getv1):
        """Test38 DockerIoAPI().search_get_page"""
        mock_searchv2.return_value = True
        mock_searchv1.return_value = False
        mock_getv2.return_value = {"count": 1, "num_pages": 1}
        mock_getv1.return_value = {"page": 1, "num_pages": 1}
        doia = DockerIoAPI(self.local)
        out = doia.search_get_page("SOMETHING")
        self.assertEqual(out, {"count": 1, "num_pages": 1})

        mock_searchv2.return_value = False
        mock_searchv1.return_value = True
        mock_getv2.return_value = {"count": 1, "num_pages": 1}
        mock_getv1.return_value = {"page": 1, "num_pages": 1}
        doia = DockerIoAPI(self.local)
        out = doia.search_get_page("SOMETHING")
        self.assertEqual(out, {"page": 1, "num_pages": 1})

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: NVIDIA mode
"""

from unittest import TestCase, main
from unittest.mock import patch, Mock
from udocker.config import Config
from udocker.engine.nvidia import NvidiaMode


class NvidiaModeTestCase(TestCase):
    """Test PRootEngine() class for containers execution."""

    def setUp(self):
        Config().getconf()
        Config().conf['nvi_dev_list'] = ['/dev/nvidia']
        self.cont_id = "12345a"
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo
        str_cd = str_local + '.cd_container'
        self.cdcont = patch(str_cd, return_value="/" + self.cont_id)
        self.mock_cdcont = self.cdcont.start()

    def tearDown(self):
        self.lrepo.stop()
        self.cdcont.stop()

    def test_01_init(self):
        """Test01 NvidiaMode() constructor."""
        cdir = "/" + self.cont_id
        nvmode = NvidiaMode(self.local, self.cont_id)
        self.assertTrue(self.mock_cdcont.called)
        self.assertEqual(nvmode.container_dir, cdir)
        self.assertEqual(nvmode.container_root, cdir + "/ROOT")
        self.assertEqual(nvmode._container_nvidia_set, cdir + "/nvidia")

    @patch('udocker.engine.nvidia.os.path.exists')
    def test_02__files_exist(self, mock_exists):
        """Test02 NvidiaMode._files_exist."""
        cont_dst_dir = "/home/.udocker/container"
        files_list = ["a", "b"]
        mock_exists.return_value = False
        nvmode = NvidiaMode(self.local, self.cont_id)
        nvmode._files_exist(cont_dst_dir, files_list)
        self.assertTrue(mock_exists.called)

    @patch('udocker.engine.nvidia.Msg')
    @patch('udocker.engine.nvidia.os.access')
    @patch('udocker.engine.nvidia.stat')
    @patch('udocker.engine.nvidia.shutil.copy2')
    @patch('udocker.engine.nvidia.os.symlink')
    @patch('udocker.engine.nvidia.os.readlink')
    @patch('udocker.engine.nvidia.os.chmod')
    @patch('udocker.engine.nvidia.os.makedirs')
    @patch('udocker.engine.nvidia.os.path.isdir')
    @patch('udocker.engine.nvidia.os.path.dirname')
    @patch('udocker.engine.nvidia.os.remove')
    @patch('udocker.engine.nvidia.os.path.islink')
    @patch('udocker.engine.nvidia.os.path.isfile')
    def test_03__copy_files(self, mock_isfile, mock_islink, mock_rm,
                            mock_dirname, mock_isdir, mock_mkdir, mock_chmod,
                            mock_readln, mock_symln, mock_copy2, mock_stat,
                            mock_access, mock_msg):
        """Test03 NvidiaMode._copy_files."""
        hsrc_dir = "/usr/lib"
        cdst_dir = "/hone/.udocker/cont/ROOT/usr/lib"
        flist = ["a"]
        force = False
        mock_msg.level = 0
        mock_isfile.side_effect = [True, False]
        mock_islink.side_effect = [True, False]
        mock_rm.return_value = None
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._copy_files(hsrc_dir, cdst_dir, flist, force)
        self.assertFalse(status)
        self.assertTrue(mock_isfile.called)

        force = True
        mock_msg.level = 0
        mock_isfile.side_effect = [True, True]
        mock_islink.side_effect = [True, False]
        mock_dirname.side_effect = [None, None]
        mock_rm.return_value = None
        mock_isdir.return_value = True
        mock_mkdir.return_value = None
        mock_chmod.side_effect = [644, 755]
        mock_readln.return_value = "/usr/xxx"
        mock_symln.return_value = None
        mock_copy2.return_value = None
        mock_stat.side_effect = [444, 222, 111, 111]
        mock_access.return_value = False
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._copy_files(hsrc_dir, cdst_dir, flist, force)
        self.assertTrue(status)
        self.assertTrue(mock_isfile.called)
        self.assertTrue(mock_rm.called)

    @patch('udocker.engine.nvidia.glob.glob')
    def test_04__get_nvidia_libs(self, mock_glob):
        """Test04 NvidiaMode._get_nvidia_libs."""
        host_dir = "/usr/lib"
        Config.conf['nvi_lib_list'] = ['/lib/libnvidia.so']
        mock_glob.return_value = ['/lib/libnvidia.so']
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._get_nvidia_libs(host_dir)
        self.assertEqual(status, Config.conf['nvi_lib_list'])

    # @patch('udocker.engine.nvidia.os.path.realpath')
    # @patch('udocker.engine.nvidia.Uprocess.get_output')
    # def test_05__find_host_dir_ldconfig(self, mock_uproc, mock_realp):
    #     """Test05 NvidiaMode._find_host_dir_ldconfig."""
    #     res_set = set()
    #     mock_uproc.return_value = \
    #       "libnvidia-cfg.so.1 (libc6,x86-64) => \
    #       /lib/x86_64-linux-gnu/libnvidia-cfg.so.1"
    #     mock_realp.return_value = "/lib/x86_64-linux-gnu/libnvidia-cfg.so.1"
    #     res_set.add("/lib/x86_64-linux-gnu/")
    #     nvmode = NvidiaMode(self.local, self.cont_id)
    #     status = nvmode._find_host_dir_ldconfig()
    #     self.assertEqual(status, res_set)

    # def test_06__find_host_dir_ldpath(self):
    #     """Test06 NvidiaMode._find_host_dir_ldpath."""

    @patch.object(NvidiaMode, '_find_host_dir_ldconfig')
    @patch.object(NvidiaMode, '_find_host_dir_ldpath')
    @patch('udocker.engine.nvidia.Msg')
    def test_07__find_host_dir(self, mock_msg, mock_ldpath, mock_ldconf):
        """Test07 NvidiaMode._find_host_dir."""
        res = set()
        mock_msg.return_value.level.return_value = 0
        mock_ldpath.side_effect = [res, res]
        mock_ldconf.return_value = res
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._find_host_dir()
        self.assertEqual(status, res)
        self.assertTrue(mock_ldpath.called)
        self.assertTrue(mock_ldconf.called)

        res = set()
        res.add("/lib/x86_64-linux-gnu/")
        mock_msg.return_value.level.return_value = 0
        mock_ldpath.side_effect = [set(), set()]
        mock_ldconf.return_value = res
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._find_host_dir()
        self.assertEqual(status, res)

    @patch('udocker.engine.nvidia.os.path.isdir')
    @patch('udocker.engine.nvidia.Msg')
    def test_08__find_cont_dir(self, mock_msg, mock_isdir):
        """Test08 NvidiaMode._find_cont_dir."""
        cdir = ""
        mock_msg.return_value.level.return_value = 0
        mock_isdir.side_effect = [False, False]
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._find_cont_dir()
        self.assertEqual(status, cdir)

        cdir = "/usr/lib/x86_64-linux-gnu"
        mock_msg.return_value.level.return_value = 0
        mock_isdir.side_effect = [True, False]
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._find_cont_dir()
        self.assertEqual(status, cdir)

    @patch.object(NvidiaMode, '_get_nvidia_libs')
    @patch.object(NvidiaMode, '_find_host_dir_ldconfig')
    @patch('udocker.engine.nvidia.Msg')
    def test_09__installation_exists(self, mock_msg, mock_ldconf,
                                     mock_libs):
        """Test09 NvidiaMode._installation_exists."""
        host_dir = "/usr/lib"
        cont_dir = "/home/.udocker/cont/usr/lib"
        mock_msg.return_value.level.return_value = 0
        mock_ldconf.side_effect = [set(), set(), set()]
        mock_libs.return_value = ['/lib/libnvidia.so']
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode._installation_exists(host_dir, cont_dir)
        self.assertFalse(status)

    @patch('udocker.engine.nvidia.FileUtil.putdata')
    @patch.object(NvidiaMode, '_copy_files')
    @patch.object(NvidiaMode, '_get_nvidia_libs')
    @patch.object(NvidiaMode, '_installation_exists')
    @patch.object(NvidiaMode, '_find_cont_dir')
    @patch.object(NvidiaMode, '_find_host_dir')
    @patch('udocker.engine.nvidia.Msg')
    def test_10_set_mode(self, mock_msg, mock_findhdir,
                         mock_findcdir, mock_instexist, mock_libs,
                         mock_cpfiles, mock_futilput):
        """Test10 NvidiaMode.set_mode()."""
        mock_msg.return_value.level.return_value = 0
        self.mock_cdcont.return_value = ""
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode.set_mode()
        self.assertTrue(mock_msg().err.called)
        self.assertFalse(status)

        self.mock_cdcont.return_value = "/home/.udocker/cont"
        mock_findhdir.return_value = set()
        mock_findcdir.return_value = "/usr/lib/x86_64-linux-gnu"
        nvmode = NvidiaMode(self.local, self.cont_id)
        nvmode.container_dir = "/" + self.cont_id
        status = nvmode.set_mode()
        self.assertTrue(mock_msg().err.called)
        self.assertFalse(status)

        self.mock_cdcont.return_value = "/home/.udocker/cont"
        host_dir = set()
        mock_findhdir.return_value = host_dir.update("/usr/lib")
        mock_findcdir.return_value = ""
        nvmode = NvidiaMode(self.local, self.cont_id)
        nvmode.container_dir = "/" + self.cont_id
        status = nvmode.set_mode()
        self.assertTrue(mock_msg().err.called)
        self.assertFalse(status)

        # TODO: need work
        self.mock_cdcont.return_value = "/home/.udocker/cont"
        host_dir = set()
        host_dir.add("/usr/lib")
        mock_findhdir.return_value = host_dir
        mock_findcdir.return_value = "/usr/lib/x86_64-linux-gnu"
        mock_libs.return_value = ['/lib/libnvidia.so']
        mock_instexist.return_value = False
        mock_cpfiles.side_effect = [True, True, True]
        mock_futilput.return_value = None
        nvmode = NvidiaMode(self.local, self.cont_id)
        nvmode.container_dir = "/" + self.cont_id
        status = nvmode.set_mode(True)
        self.assertTrue(mock_findhdir.called)
        self.assertTrue(mock_findcdir.called)
        # self.assertTrue(mock_instexist.called)
        # self.assertTrue(mock_futilput.called)
        # self.assertTrue(status)

    @patch('udocker.engine.nvidia.os.path.exists')
    def test_11_get_mode(self, mock_exists):
        """Test11 NvidiaMode.get_mode()."""
        mock_exists.return_value = True
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode.get_mode()
        self.assertTrue(mock_exists.called)
        self.assertTrue(status)

    @patch('udocker.engine.nvidia.glob.glob')
    def test_12_get_devices(self, mock_glob):
        """Test12 NvidiaMode.get_devices()."""
        mock_glob.return_value = ['/dev/nvidia']
        nvmode = NvidiaMode(self.local, self.cont_id)
        status = nvmode.get_devices()
        self.assertEqual(status, Config().conf['nvi_dev_list'])


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: ChkSUM
"""

from unittest import TestCase, main
from unittest.mock import patch, Mock, mock_open
from io import BytesIO as strio
from udocker.utils.chksum import ChkSUM

BUILTINS = "builtins"


class ChkSUMTestCase(TestCase):
    """Test ChkSUM()."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @patch('udocker.utils.chksum.hashlib.sha512')
    @patch('udocker.utils.chksum.hashlib.sha256')
    @patch.object(ChkSUM, '_openssl_sha512', autospec=True)
    @patch.object(ChkSUM, '_hashlib_sha512', autospec=True)
    @patch.object(ChkSUM, '_openssl_sha256', autospec=True)
    @patch.object(ChkSUM, '_hashlib_sha256', autospec=True)
    def test_01__init(self, mock_hash256, mock_ossl256,
                      mock_hash512, mock_ossl512,
                      mock_lib256, mock_lib512):
        """Test01 ChkSUM() constructor."""
        mock_hash256.return_value = '123456'
        mock_hash512.return_value = '1234567891011'
        mock_lib256.return_value = True
        mock_lib512.return_value = True
        chksum = ChkSUM()
        self.assertTrue(mock_lib256.called)
        self.assertTrue(mock_lib512.called)
        # self.assertTrue(mock_hash256.called)
        # self.assertTrue(mock_hash512.called)

        # with self.assertRaises(NameError):
        #     ChkSUM()
        #     self.assertTrue(mock_ossl256.called)


    # def test_02__hashlib(self):
    #     """Test02 ChkSUM()._hashlib."""

    # def test_03__hashlib_sha256(self):
    #     """Test03 ChkSUM()._hashlib_sha256."""

    # def test_04__hashlib_sha512(self):
    #     """Test04 ChkSUM()._hashlib_sha512."""

    # def test_05__openssl(self):
    #     """Test05 ChkSUM()._openssl."""

    # def test_06__openssl_sha256(self):
    #     """Test06 ChkSUM()._openssl_sha256."""

    # def test_07__openssl_sha512(self):
    #     """Test07 ChkSUM()._openssl_sha512."""

    # def test_08_sha256(self):
    #     """Test08 ChkSUM().sha256."""

    # def test_09_sha512(self):
    #     """Test09 ChkSUM().sha512."""

    # def test_10_hash(self):
    #     """Test10 ChkSUM().hash."""
    #     sha256sum = (
    #       "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855")
    #     cksum = ChkSUM()
    #     file_data = strio('qwerty')
    #     with patch(BUILTINS + '.open', mock_open()) as mopen:
    #         mopen.return_value.__iter__ = (
    #             lambda self: iter(file_data.readline, ''))
    #         status = cksum.hash("filename", "sha256")
    #         self.assertEqual(status, sha256sum)

    #     sha512sum = ("cf83e1357eefb8bdf1542850d66d8007d620e4050b"
    #                  "5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff831"
    #                  "8d2877eec2f63b931bd47417a81a538327af927da3e")
    #     cksum = ChkSUM()
    #     file_data = strio('qwerty')
    #     with patch(BUILTINS + '.open', mock_open()) as mopen:
    #         mopen.return_value.__iter__ = (
    #             lambda self: iter(file_data.readline, ''))
    #         status = cksum.hash("filename", "sha512")
    #         self.assertEqual(status, sha512sum)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Unique IDentifiers"""

import os
import time
import string

try:
    import random
except ImportError:
    pass

try:
    import uuid
except ImportError:
    pass


class Unique(object):
    """Produce unique identifiers for container names, temporary
    file names and other purposes. If module uuid does not exist
    it tries to use as last option the random generator.
    """
    def __init__(self):
        self.string_set = "abcdef"
        self.def_name = "udocker"

    def _rnd(self, size):
        """Generate a random string"""
        return "".join(
            random.sample(self.string_set * 64 + string.digits * 64, size))

    def uuid(self, name):
        """Get an ID"""
        if not name:
            name = self.def_name
        try:
            return str(uuid.uuid3(uuid.uuid4(), str(name) + str(time.time())))
        except (NameError, AttributeError):
            return f"{self._rnd(8)}-{self._rnd(4)}-{self._rnd(4)}-{self._rnd(4)}-{self._rnd(12)}"

    def imagename(self):
        """Get a container image name"""
        return self._rnd(16)

    def imagetag(self):
        """Get a container image tag"""
        return self._rnd(10)

    def layer_v1(self):
        """Get a random container layer name"""
        return self._rnd(64)

    def filename(self, filename):
        """Get a filename"""
        prefix = self.def_name + '-' + str(os.getpid()) + '-'
        try:
            return (prefix +
                    str(uuid.uuid3(uuid.uuid4(), str(time.time()))) +
                    '-' + str(filename))
        except (NameError, AttributeError):
            return prefix + self.uuid(filename) + '-' + str(filename)



# -*- coding: utf-8 -*-
"""Information about the host or guest/container"""

import os
import re

from udocker.utils.uprocess import Uprocess
from udocker.utils.fileutil import FileUtil


class OSInfo(object):
    """Get os information from a directory tree"""

    _binarylist = ["/lib64/ld-linux-x86-64.so",
                   "/lib64/ld-linux-x86-64.so.2",
                   "/lib64/ld-linux-x86-64.so.3",
                   "/bin/bash", "/bin/sh", "/bin/zsh",
                   "/bin/csh", "/bin/tcsh", "/bin/ash",
                   "/bin/ls", "/bin/busybox",
                   "/system/bin/sh", "/system/bin/ls",
                   "/lib/ld-linux.so",
                   "/lib/ld-linux.so.2",
                  ]

    def __init__(self, root_dir):
        self._root_dir = root_dir

    def get_filetype(self, filename):
        """Get the file architecture"""
        filetype = ""
        if not filename.startswith(self._root_dir):
            filename = self._root_dir + '/' + filename
        if os.path.islink(filename):
            f_path = os.readlink(filename)
            if not f_path.startswith('/'):
                f_path = os.path.dirname(filename) + '/' + f_path
            return self.get_filetype(f_path)
        if os.path.isfile(filename):
            filetype = Uprocess().get_output(["file", filename])
            if not filetype:
                filetype = Uprocess().get_output(["readelf", "-h", filename])
        return filetype

    def arch(self):
        """Get guest system architecture"""
        for filename in OSInfo._binarylist:
            f_path = self._root_dir + filename
            filetype = self.get_filetype(f_path)
            if not filetype:
                continue
            if "x86-64" in filetype.lower():
                return "amd64"
            if "Intel 80386" in filetype:
                return "i386"
            if "aarch64" in filetype.lower():
                return "arm64"
            if " ARM" in filetype:
                return "arm"
        return ""

    def osdistribution(self):
        """Get guest operating system distribution"""
        for f_path in FileUtil(self._root_dir + "/etc/.+-release").match():
            if os.path.exists(f_path):
                osinfo = FileUtil(f_path).getdata('r')
                match = re.match(r"([^=]+) release (\d+)", osinfo)
                if match and match.group(1):
                    return (match.group(1).split(' ')[0],
                            match.group(2).split('.')[0])
        f_path = self._root_dir + "/etc/lsb-release"
        if os.path.exists(f_path):
            distribution = ""
            version = ""
            osinfo = FileUtil(f_path).getdata('r')
            match = re.search(r"DISTRIB_ID=(.+)(\n|$)",
                              osinfo, re.MULTILINE)
            if match:
                distribution = match.group(1).split(' ')[0]
            match = re.search(r"DISTRIB_RELEASE=(.+)(\n|$)",
                              osinfo, re.MULTILINE)
            if match:
                version = match.group(1).split('.')[0]
            if distribution and version:
                return (distribution, version)
        f_path = self._root_dir + "/etc/os-release"
        if os.path.exists(f_path):
            distribution = ""
            version = ""
            osinfo = FileUtil(f_path).getdata('r')
            match = re.search(r"NAME=\"?([^ \n\"\.]+).*\"?(\n|$)",
                              osinfo, re.MULTILINE)
            if match:
                distribution = match.group(1).split(' ')[0]
            match = re.search(r"VERSION_ID=\"?([^ \n\"\.]+).*\"?(\n|$)",
                              osinfo, re.MULTILINE)
            if match:
                version = match.group(1).split('.')[0]
            if distribution and version:
                return (distribution, version)
        return ("", "")

    def osversion(self):
        """Get guest operating system"""
        if self.osdistribution()[0]:
            return "linux"
        return ""



# -*- coding: utf-8 -*-
"""fakechroot execution engine"""

import os
import sys
import re
import subprocess

from udocker import is_genstr
from udocker.engine.base import ExecutionEngineCommon
from udocker.helper.osinfo import OSInfo
from udocker.msg import Msg
from udocker.config import Config
from udocker.utils.fileutil import FileUtil
from udocker.utils.uvolume import Uvolume
from udocker.helper.elfpatcher import ElfPatcher


class FakechrootEngine(ExecutionEngineCommon):
    """Docker container execution engine using Fakechroot
    Provides a chroot like environment to run containers.
    Uses Fakechroot as chroot alternative.
    Inherits from ContainerEngine class
    """

    def __init__(self, localrepo, exec_mode):
        super(FakechrootEngine, self).__init__(localrepo, exec_mode)
        self._fakechroot_so = ""
        self._elfpatcher = None
        self._recommend_expand_symlinks = False

    def select_fakechroot_so(self):
        """Select fakechroot sharable object library"""
        image_list = []
        if Config.conf['fakechroot_so']:
            if isinstance(Config.conf['fakechroot_so'], list):
                image_list = Config.conf['fakechroot_so']
            elif is_genstr(Config.conf['fakechroot_so']):
                image_list = [Config.conf['fakechroot_so'], ]

            if "/" in Config.conf['fakechroot_so']:
                if os.path.exists(Config.conf['fakechroot_so']):
                    return os.path.realpath(Config.conf['fakechroot_so'])
        elif os.path.exists(self.container_dir + "/libfakechroot.so"):
            return self.container_dir + "/libfakechroot.so"
        else:
            lib = "libfakechroot"
            deflib = "libfakechroot.so"
            image_list = [deflib, ]
            guest = OSInfo(self.container_root)
            arch = guest.arch()
            (distro, version) = guest.osdistribution()
            if "Alpine" not in distro:
                version = version.split(".")[0]

            if arch == "amd64":
                image_list = [f"{lib}-{distro}-{version}-x86_64.so",
                              f"{lib}-{distro}-x86_64.so",
                              f"{lib}-x86_64.so", deflib]
            elif arch == "i386":
                image_list = [f"{lib}-{distro}-{version}-x86.so",
                              f"{lib}-{distro}-x86.so",
                              f"{lib}-x86.so", deflib]
            elif arch == "arm64":
                image_list = [f"{lib}-{distro}-{version}-arm64.so",
                              f"{lib}-{distro}-arm64.so",
                              f"{lib}-arm64.so", deflib]
            elif arch == "arm":
                image_list = [f"{lib}-{distro}-{version}-arm.so",
                              f"{lib}-{distro}-arm.so",
                              f"{lib}-arm.so", deflib]

        f_util = FileUtil(self.localrepo.libdir)
        fakechroot_so = f_util.find_file_in_dir(image_list)
        if not os.path.exists(fakechroot_so):
            Msg().err("Error: no libfakechroot found", image_list)
            sys.exit(1)

        Msg().out("Info: fakechroot_so:", fakechroot_so, l=Msg.DBG)
        return fakechroot_so

    def _setup_container_user(self, user):
        """Override of _setup_container_user()"""
        return self._setup_container_user_noroot(user)

    def _uid_check(self):
        """Check the uid_map string for container run command"""
        if ("user" in self.opt and (self.opt["user"] == '0' or
                                    self.opt["user"] == "root")):
            Msg().out("Warning: this engine does not support execution as root",
                      l=Msg.WAR)

    def _get_volume_bindings(self):
        """Get the volume bindings string for fakechroot run"""
        host_volumes_list = []
        map_volumes_list = []
        map_volumes_dict = {}
        for vol in self.opt["vol"]:
            (host_path, cont_path) = Uvolume(vol).split()
            if not (host_path and cont_path):
                continue
            real_host_path = os.path.realpath(host_path)
            if (host_path == cont_path and
                    Config.conf['fakechroot_expand_symlinks'] is False):
                host_volumes_list.append(host_path)
            elif (host_path == cont_path and
                  host_path in Config.conf['sysdirs_list']):
                host_volumes_list.append(host_path)
            elif host_path == cont_path and not os.path.isdir(real_host_path):
                host_volumes_list.append(host_path)
            else:
                map_volumes_dict[cont_path] = real_host_path + '!' + cont_path
                if host_path != real_host_path or os.path.isdir(real_host_path):
                    self._recommend_expand_symlinks = True
        for cont_path in sorted(map_volumes_dict, reverse=True):
            map_volumes_list.append(map_volumes_dict[cont_path])
        return (':'.join(host_volumes_list), ':'.join(map_volumes_list))

    def _get_access_filesok(self):
        """
        Circunvent mpi init issues when calling access()
        A list of certain existing files is provided
        """
        file_list = []
        for c_path in Config.conf['access_files']:
            h_file = FileUtil(self.container_root).cont2host(c_path,
                                                             self.opt["vol"])
            if h_file and os.path.exists(h_file):
                file_list.append(c_path)
        return ":".join(file_list)

    def _fakechroot_env_set(self):
        """fakechroot environment variables to set"""
        (host_volumes, map_volumes) = self._get_volume_bindings()
        self._fakechroot_so = self.select_fakechroot_so()
        access_filesok = self._get_access_filesok()
        self.opt["env"].append("PWD=" + self.opt["cwd"])
        self.opt["env"].append("FAKECHROOT_BASE=" +
                               os.path.realpath(self.container_root))
        self.opt["env"].append("LD_PRELOAD=" + self._fakechroot_so)
        if Config.conf['fakechroot_expand_symlinks'] is None:
            self.opt["env"].append("FAKECHROOT_EXPAND_SYMLINKS=" + \
                    str(self._recommend_expand_symlinks).lower())
        else:
            self.opt["env"].append("FAKECHROOT_EXPAND_SYMLINKS=" + \
                    str(Config.conf['fakechroot_expand_symlinks']).lower())

        if not self._is_volume("/tmp"):
            self.opt["env"].append("FAKECHROOT_AF_UNIX_PATH=" +
                                   Config.conf['tmpdir'])

        if host_volumes:
            self.opt["env"].append("FAKECHROOT_EXCLUDE_PATH=" + host_volumes)

        if map_volumes:
            self.opt["env"].append("FAKECHROOT_DIR_MAP=" + map_volumes)

        if Msg.level >= Msg.DBG:
            self.opt["env"].append("FAKECHROOT_DEBUG=true")
            self.opt["env"].append("LD_DEBUG=libs:files")

        if access_filesok:
            self.opt["env"].append("FAKECHROOT_ACCESS_FILESOK=" +
                                   access_filesok)

        # execution mode
        ld_library_real = self._elfpatcher.get_ld_library_path()
        xmode = self.exec_mode.get_mode()
        if xmode == "F1":
            self.opt["env"].append("FAKECHROOT_ELFLOADER=" +
                                   self._elfpatcher.get_container_loader())
            self.opt["env"].append("LD_LIBRARY_PATH=" + ld_library_real)
        elif xmode == "F2":
            self.opt["env"].append("FAKECHROOT_ELFLOADER=" +
                                   self._elfpatcher.get_container_loader())
            self.opt["env"].append("FAKECHROOT_LIBRARY_ORIG=" + ld_library_real)
            self.opt["env"].append("LD_LIBRARY_REAL=" + ld_library_real)
            self.opt["env"].append("LD_LIBRARY_PATH=" + ld_library_real)
        elif xmode == "F3":
            self.opt["env"].append("FAKECHROOT_LIBRARY_ORIG=" + ld_library_real)
            self.opt["env"].append("LD_LIBRARY_REAL=" + ld_library_real)
            self.opt["env"].append("LD_LIBRARY_PATH=" + ld_library_real)
        elif xmode == "F4":
            self.opt["env"].append("FAKECHROOT_LIBRARY_ORIG=" + ld_library_real)
            self.opt["env"].append("LD_LIBRARY_REAL=" + ld_library_real)
            self.opt["env"].append("LD_LIBRARY_PATH=" + ld_library_real)
            patchelf_exec = self._elfpatcher.select_patchelf()
            if patchelf_exec:
                self.opt["env"].append("FAKECHROOT_PATCH_PATCHELF=" +
                                       patchelf_exec)
                self.opt["env"].append("FAKECHROOT_PATCH_ELFLOADER=" +
                                       self._elfpatcher.get_container_loader())
                self.opt["env"].append("FAKECHROOT_PATCH_LAST_TIME=" +
                                       self._elfpatcher.get_patch_last_time())

    def _run_invalid_options(self):
        """check -p --publish -P --publish-all --net-coop"""
        if self.opt["portsmap"]:
            Msg().out("Warning: this execution mode does not support "
                      "-p --publish", l=Msg.WAR)
        if self.opt["netcoop"]:
            Msg().out("Warning: this execution mode does not support "
                      "-P --netcoop --publish-all", l=Msg.WAR)

    def _run_add_script_support(self, exec_path):
        """Add an interpreter for non binary executables (scripts)"""
        filetype = OSInfo(self.container_root).get_filetype(exec_path)
        if "ELF" in filetype and ("static" in filetype or
                                  "dynamic" in filetype):
            self.opt["cmd"][0] = exec_path
            return []
        env_exec = FileUtil("env").find_exec("/bin:/usr/bin",
                                             self.container_root)
        if  env_exec:
            return [self.container_root + '/' + env_exec, ]
        relc_path = exec_path.split(self.container_root, 1)[-1]
        real_path = FileUtil(self.container_root).cont2host(relc_path,
                                                            self.opt["vol"])
        hashbang = FileUtil(real_path).get1stline()
        match = re.match("#! *([^ ]+)(.*)", hashbang)
        if match and not match.group(1).startswith('/'):
            Msg().err("Error: no such file", match.group(1), "in", exec_path)
            sys.exit(1)
        elif match:
            interpreter = [self.container_root + '/' + match.group(1), ]
            if match.group(2):
                interpreter.extend(match.group(2).strip().split(' '))
            self.opt["cmd"][0] = exec_path.split(self.container_root, 1)[-1]
            return interpreter
        sh_exec = FileUtil("sh").find_exec(self.opt["env"].getenv("PATH"),
                                           self.container_root)
        if sh_exec:
            return [self.container_root + '/' + sh_exec, ]
        Msg().err("Error: sh not found")
        sys.exit(1)

    def run(self, container_id):
        """Execute a Docker container using Fakechroot. This is the main
        method invoked to run the a container with Fakechroot.
          * argument: container_id or name
          * options:  many via self.opt see the help
        """

        # warning root execution not supported
        self._uid_check()

        # setup execution
        exec_path = self._run_init(container_id)
        if not exec_path:
            return 2

        self._run_invalid_options()

        # execution mode and get patcher
        xmode = self.exec_mode.get_mode()
        self._elfpatcher = ElfPatcher(self.localrepo, self.container_id)

        # verify if container pathnames are correct for this mode
        if not self._elfpatcher.check_container_path():
            Msg().out("Warning: container path mismatch, use setup to convert",
                      l=Msg.WAR)

        # set basic environment variables
        self._run_env_set()
        self._fakechroot_env_set()

        # if not --hostenv clean the environment
        self._run_env_cleanup_list()

        # build the actual command
        cmd_l = self._set_cpu_affinity()
        cmd_l.extend(["env", "-i", ])
        cmd_l.extend(self.opt["env"].list())
        if xmode in ("F1", "F2"):
            container_loader = self._elfpatcher.get_container_loader()
            if container_loader:
                cmd_l.append(container_loader)
        cmd_l.extend(self._run_add_script_support(exec_path))
        cmd_l.extend(self.opt["cmd"])
        Msg().out("CMD =", cmd_l, l=Msg.VER)

        # execute
        self._run_banner(self.opt["cmd"][0], '#')
        cwd = FileUtil(self.container_root).cont2host(self.opt["cwd"],
                                                      self.opt["vol"])
        status = subprocess.call(cmd_l, shell=False, close_fds=True, cwd=cwd)
        return status



#!/usr/bin/env python
"""
udocker unit tests: OciLocalFileAPI
"""

from unittest import TestCase, main
from unittest.mock import patch, Mock
from udocker.oci import OciLocalFileAPI


class OciLocalFileAPITestCase(TestCase):
    """Test OciLocalFileAPI()."""

    def setUp(self):
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    # def test_01__init(self):
    #     """Test01 OciLocalFileAPI() constructor."""

    @patch('udocker.oci.FileUtil.isdir')
    @patch('udocker.oci.os.listdir')
    @patch('udocker.container.localrepo.LocalRepository.load_json')
    def test_02__load_structure(self, mock_ljson, mock_oslist, mock_isdir):
        """Test02 OciLocalFileAPI()._load_structure."""
        mock_ljson.side_effect = [[], []]
        status = OciLocalFileAPI(self.local)._load_structure('tmpimg')
        self.assertEqual(status, {})

        out_res = {'repolayers': {},
                   'manifest': {},
                   'oci-layout': 'oci_lay1',
                   'index': 'idx1'}
        mock_ljson.side_effect = ['oci_lay1', 'idx1']
        mock_oslist.return_value = ['f1']
        mock_isdir.return_value = False
        status = OciLocalFileAPI(self.local)._load_structure('tmpimg')
        self.assertEqual(status, out_res)

        out_res = {'repolayers': {'f1:f2': {'layer_a': 'f1',
                                            'layer_f': 'tmpimg/blobs/f1/f2',
                                            'layer_h': 'f2'}},
                   'manifest': {},
                   'oci-layout': 'oci_lay1',
                   'index': 'idx1'}
        mock_ljson.side_effect = ['oci_lay1', 'idx1']
        mock_oslist.side_effect = [['f1'], ['f2']]
        mock_isdir.return_value = True
        status = OciLocalFileAPI(self.local)._load_structure('tmpimg')
        self.assertEqual(status, out_res)

    def test_03__get_from_manifest(self):
        """Test03 OciLocalFileAPI()._get_from_manifest."""
        imgtag = '123'
        struct = {'manifest': {'123': {'json': {'layers': [{'digest': 'd1'},
                                                           {'digest': 'd2'}],
                                                'config': {'digest': 'dgt'}}}}}
        lay_out = ['d2', 'd1']
        conf_out = 'dgt'
        status = OciLocalFileAPI(self.local)._get_from_manifest(struct, imgtag)
        self.assertEqual(status, (conf_out, lay_out))

    # @patch('udocker.oci.Unique.imagename')
    # @patch('udocker.oci.Unique.imagetag')
    # @patch('udocker.container.localrepo.LocalRepository.load_json')
    # def test_04__load_manifest(self, mock_ljson, mock_uniqtag, mock_uniqname):
    #     """Test04 OciLocalFileAPI()._load_manifest."""
    #     manifest = {'annotations':
    #           {'org.opencontainers.image.ref.name': '123'}}
    #     mock_uniqtag.return_value = '123'
    #     mock_uniqname.return_value = 'imgname'
    #     mock_ljson.return_value = {'layers': [{'digest': 'd1'},
    #                                           {'digest': 'd2'}],
    #                                'config': {'digest': 'dgt'}}
    #     struct = {'manifest': {'123': {'json': mock_ljson}},
    #               'repolayers': {'digest': {'layer_a': 'f1',
    #                                         'layer_f': 'tmpimg/blobs/f1/f2',
    #                                         'layer_h': 'f2'}}
    #              }
    #     status = OciLocalFileAPI(self.local)._load_manifest(struct, manifest)
    #     self.assertEqual(status, ['123'])

    # def test_05__load_repositories(self):
    #     """Test05 OciLocalFileAPI()._load_repositories."""

    # def test_06__load_image_step2(self):
    #     """Test07 OciLocalFileAPI()._load_image_step2."""

    @patch('udocker.oci.Msg.err')
    @patch.object(OciLocalFileAPI, '_load_repositories')
    @patch.object(OciLocalFileAPI, '_load_structure')
    def test_07_load(self, mock_loadstruct, mock_loadrepo, mock_msg):
        """Test07 OciLocalFileAPI().load."""
        tmpdir = '/ROOT'
        imgrepo = 'somerepo'
        mock_loadstruct.return_value = {}
        status = OciLocalFileAPI(self.local).load(tmpdir, imgrepo)
        self.assertTrue(mock_msg.called)
        self.assertEqual(status, [])

        tmpdir = '/ROOT'
        imgrepo = 'somerepo'
        mock_loadstruct.return_value = {'repolayers':
                                            {'f1:f2': {'layer_a': 'f1',
                                             'layer_f': 'tmpimg/blobs/f1/f2',
                                             'layer_h': 'f2'}},
                                        'manifest': {},
                                        'oci-layout': 'oci_lay1',
                                        'index': 'idx1'}
        mock_loadrepo.return_value = ['r1', 'r2']
        status = OciLocalFileAPI(self.local).load(tmpdir, imgrepo)
        self.assertEqual(status, ['r1', 'r2'])


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Management and tools for the keystore"""

import os
import json

from udocker.utils.fileutil import FileUtil
from udocker.helper.hostinfo import HostInfo


class KeyStore(object):
    """Basic storage for authentication tokens to be used
    with dockerhub and private repositories
    """

    def __init__(self, keystore_file):
        """Initialize keystone"""
        self.keystore_file = keystore_file
        self.credential = {}

    def _verify_keystore(self):
        """Verify the keystore file and directory"""
        keystore_uid = FileUtil(self.keystore_file).uid()
        if keystore_uid not in (-1, HostInfo.uid):
            raise IOError(f"not owner of keystore: {self.keystore_file}")
        keystore_dir = os.path.dirname(self.keystore_file)
        if FileUtil(keystore_dir).uid() != HostInfo.uid:
            raise IOError(f"keystore dir not found or not owner: {keystore_dir}")
        if (keystore_uid != -1 and (os.stat(self.keystore_file).st_mode & 0o077)):
            raise IOError(f"keystore is accessible to group or others: {self.keystore_file}")

    def _read_all(self):
        """Read all credentials from file"""
        try:
            with open(self.keystore_file, "r", encoding='utf-8') as filep:
                return json.load(filep)
        except (IOError, OSError, ValueError):
            return {}

    def _shred(self):
        """Shred file content"""
        exit_status = 0
        self._verify_keystore()
        try:
            size = FileUtil(self.keystore_file).size()
            with open(self.keystore_file, "rb+") as filep:
                filep.write(b" " * size)
        except (IOError, OSError):
            exit_status = 1
            return exit_status
        return exit_status

    def _write_all(self, auths):
        """Write all credentials to file"""
        exit_status = 0
        self._verify_keystore()
        oldmask = None
        try:
            oldmask = os.umask(0o77)
            with open(self.keystore_file, "w", encoding='utf-8') as filep:
                json.dump(auths, filep)
            os.umask(oldmask)
        except (IOError, OSError):
            if oldmask is not None:
                os.umask(oldmask)
            exit_status = 1
            return exit_status
        return exit_status

    def get(self, url):
        """Get credential from keystore for given url"""
        auths = self._read_all()
        try:
            self.credential = auths[url]
            return self.credential["auth"]
        except KeyError:
            pass
        return ""

    def put(self, url, credential, email):
        """Put credential in keystore for given url"""
        if not credential:
            return 1
        auths = self._read_all()
        auths[url] = {"auth": credential, "email": email, }
        self._shred()
        return self._write_all(auths)

    def delete(self, url):
        """Delete credential from keystore for given url"""
        self._verify_keystore()
        auths = self._read_all()
        try:
            del auths[url]
        except KeyError:
            return 1
        self._shred()
        return self._write_all(auths)

    def erase(self):
        """Delete all credentials from keystore"""
        self._verify_keystore()
        try:
            self._shred()
            os.unlink(self.keystore_file)
        except (IOError, OSError):
            return 1
        return 0



# -*- coding: utf-8 -*-
"""Local repository management for images, containers"""

import os
import re
import sys
import stat
import json

from udocker import is_genstr
from udocker.config import Config
from udocker.msg import Msg
from udocker.utils.fileutil import FileUtil
from udocker.utils.chksum import ChkSUM
from udocker.utils.uprocess import Uprocess
from udocker.helper.osinfo import OSInfo


class LocalRepository(object):
    """Implements a basic repository for images and containers.
    The repository will be usually in the user home directory.
    The repository has a simple directory structure:
    1. layers    : one dir containing all image layers so that
                   layers shared among images are not duplicated
    2. containers: has inside one directory per container,
                   each dir has a ROOT with the extracted image
    3. repos:      has inside a directory tree of repos the
                   leaf repo dirs are called tags and contain the
                   image data (these are links both to layer tarballs
                   and json metadata files.
    4. bin:        contains executables (PRoot)
    5. lib:        contains python libraries
    """

    def __init__(self, topdir=None):
        self.topdir = topdir if topdir else Config.conf['topdir']
        self.bindir = Config.conf['bindir']
        self.libdir = Config.conf['libdir']
        self.docdir = Config.conf['docdir']
        self.reposdir = Config.conf['reposdir']
        self.layersdir = Config.conf['layersdir']
        self.containersdir = Config.conf['containersdir']
        self.homedir = Config.conf['homedir']

        if not self.bindir:
            self.bindir = self.topdir + "/bin"
        if not self.libdir:
            self.libdir = self.topdir + "/lib"
        if not self.docdir:
            self.docdir = self.topdir + "/doc"
        if not self.reposdir:
            self.reposdir = self.topdir + "/repos"
        if not self.layersdir:
            self.layersdir = self.topdir + "/layers"
        if not self.containersdir:
            self.containersdir = self.topdir + "/containers"

        self.cur_repodir = ""
        self.cur_tagdir = ""
        self.cur_containerdir = ""

        FileUtil(self.reposdir).register_prefix()
        FileUtil(self.layersdir).register_prefix()
        FileUtil(self.containersdir).register_prefix()

    def setup(self, topdir=None):
        """change to a different localrepo"""
        self.__init__(topdir)

    def create_repo(self):
        """creates properties with pathnames for easy
        access to the several repository directories
        """
        try:
            if not os.path.exists(self.topdir):
                os.makedirs(self.topdir)
            if not os.path.exists(self.reposdir):
                os.makedirs(self.reposdir)
            if not os.path.exists(self.layersdir):
                os.makedirs(self.layersdir)
            if not os.path.exists(self.containersdir):
                os.makedirs(self.containersdir)
            if not os.path.exists(self.bindir):
                os.makedirs(self.bindir)
            if not os.path.exists(self.libdir):
                os.makedirs(self.libdir)
            if not os.path.exists(self.docdir):
                os.makedirs(self.docdir)
            if not (Config.conf['keystore'].startswith("/") or \
                    os.path.exists(self.homedir)):
                os.makedirs(self.homedir)
        except(IOError, OSError):
            return False
        return True

    def is_repo(self):
        """check if directory structure corresponds to a repo"""
        dirs_exist = [os.path.exists(self.reposdir),
                      os.path.exists(self.layersdir),
                      os.path.exists(self.containersdir),
                      os.path.exists(self.bindir),
                      os.path.exists(self.libdir)]
        return all(dirs_exist)

    def is_container_id(self, obj):
        """Verify if the provided object matches the format of a
        local container id.
        """
        if not is_genstr(obj):
            return False
        match = re.match(
            "^[a-z0-9]+-[a-z0-9]+-[a-z0-9]+-[a-z0-9]+-[a-z0-9]+$", obj)
        if match:
            return True
        return False

    def protect_container(self, container_id):
        """Protect a container directory against deletion"""
        return self._protect(self.cd_container(container_id))

    def unprotect_container(self, container_id):
        """Remove the protection against deletion"""
        return self._unprotect(self.cd_container(container_id))

    def isprotected_container(self, container_id):
        """See if a container directory is protected"""
        return self._isprotected(self.cd_container(container_id))

    def _protect(self, directory):
        """Set the protection mark in a container or image tag"""
        try:
            # touch create version file
            open(directory + "/PROTECT", 'w', encoding='utf-8').close()
            return True
        except (IOError, OSError):
            return False

    def _unprotect(self, directory):
        """Remove protection mark from container or image tag"""
        return FileUtil(directory + "/PROTECT").remove()

    def _isprotected(self, directory):
        """See if container or image tag are protected"""
        return os.path.exists(directory + "/PROTECT")

    def iswriteable_container(self, container_id):
        """See if a container root dir is writable by this user"""
        container_root = self.cd_container(container_id) + "/ROOT"
        if not os.path.exists(container_root):
            return 2
        if not os.path.isdir(container_root):
            return 3
        if os.access(container_root, os.W_OK):
            return 1
        return 0

    def get_size(self, container_id):
        """See if a container root dir is writable by this user"""
        container_root = self.cd_container(container_id) + "/ROOT"
        try:
            size, dummy = Uprocess().get_output(["du", "-s", "-m", "-x",
                                                 container_root]).split()
            return int(size)
        except (ValueError, NameError, AttributeError):
            return -1

    def get_containers_list(self, dir_only=True):
        """Get a list of all containers in the local repo
        dir_only: is optional and indicates
                  if True a summary list of container_ids and names
                  if False  an extended listing containing further
                  container information
        """
        containers_list = []
        if not os.path.isdir(self.containersdir):
            return []
        for fname in os.listdir(self.containersdir):
            container_dir = self.containersdir + '/' + fname
            if os.path.isdir(container_dir):
                try:
                    filep = open(container_dir + "/imagerepo.name", 'r',
                                 encoding='utf-8')
                except (IOError, OSError):
                    reponame = ""
                else:
                    reponame = filep.read()
                    filep.close()
                if dir_only:
                    containers_list.append(container_dir)
                elif not os.path.islink(container_dir):
                    names = self.get_container_name(fname)
                    if not names:
                        names = ""
                    containers_list.append((fname, reponame, str(names)))
        return containers_list

    def del_container(self, container_id, force=False):
        """Delete a container tree, the image layers are untouched"""
        container_dir = self.cd_container(container_id)
        if not container_dir:
            return False

        if container_dir in self.get_containers_list(True):
            for name in self.get_container_name(container_id):
                self.del_container_name(name)  # delete aliases links
            if force:
                FileUtil(container_dir).rchmod(stat.S_IWUSR | stat.S_IRUSR,
                                                stat.S_IWUSR | stat.S_IRUSR |
                                                stat.S_IXUSR)
            if FileUtil(container_dir).remove(recursive=True):
                self.cur_containerdir = ""
                return True

        return False

    def cd_container(self, container_id):
        """Select a container directory for further operations"""
        container_dir = self.containersdir + '/' + str(container_id)
        if os.path.exists(container_dir):
            if container_dir in self.get_containers_list(True):
                return container_dir
        return ""

    def _symlink(self, existing_file, link_file):
        """Create relative symbolic links"""
        if os.path.exists(link_file):
            return False
        rel_path_to_existing = os.path.relpath(
            existing_file, os.path.dirname(link_file))
        try:
            os.symlink(rel_path_to_existing, link_file)
        except (IOError, OSError):
            return False
        return True

    def _name_is_valid(self, name):
        """Check name alias validity"""
        invalid_chars = ("/", ".", " ", "[", "]")
        if name and any(x in name for x in invalid_chars):
            return False
        return not len(name) > 2048

    def set_container_name(self, container_id, name):
        """Associates a name to a container id The container can
        then be referenced either by its id or by its name.
        """
        if self._name_is_valid(name):
            container_dir = self.cd_container(container_id)
            if container_dir:
                linkname = os.path.realpath(self.containersdir + '/' + name)
                if os.path.exists(linkname):
                    return False
                real_container_dir = os.path.realpath(container_dir)
                return self._symlink(real_container_dir, linkname)
        return False

    def del_container_name(self, name):
        """Remove a name previously associated to a container"""
        if self._name_is_valid(name):
            linkname = self.containersdir + '/' + name
            if os.path.islink(linkname):
                return FileUtil(linkname).remove()
        return False

    def get_container_id(self, container_name):
        """From a container name obtain its container_id"""
        if container_name:
            pathname = self.containersdir + "/" + container_name
            if os.path.islink(pathname):
                return os.path.basename(os.readlink(pathname))
            if os.path.isdir(pathname):
                return container_name

        return ""

    def get_container_name(self, container_id):
        """From a container_id obtain its name(s)"""
        if not os.path.isdir(self.containersdir):
            return []
        link_list = []
        for fname in os.listdir(self.containersdir):
            container = self.containersdir + "/" + fname
            if os.path.islink(container):
                real_container = os.readlink(container)
                if os.path.basename(real_container) == container_id:
                    link_list.append(fname)
        return link_list

    def setup_container(self, imagerepo, tag, container_id):
        """Create the directory structure for a container"""
        container_dir = self.containersdir + "/" + str(container_id)
        if os.path.exists(container_dir):
            return ""
        try:
            os.makedirs(container_dir + "/ROOT")
            out_imagerepo = open(container_dir + "/imagerepo.name", 'w',
                                 encoding='utf-8')
        except (IOError, OSError):
            return None
        else:
            out_imagerepo.write(imagerepo + ":" + tag)
            out_imagerepo.close()
            self.cur_containerdir = container_dir
            return container_dir

    def _is_tag(self, tag_dir):
        """Does this directory contain an image tag ?
        An image TAG indicates that this repo directory
        contains references to layers and metadata from
        which we can extract a container.
        """
        try:
            if os.path.isfile(tag_dir + "/TAG"):
                return True
        except (IOError, OSError):
            pass
        return False

    def protect_imagerepo(self, imagerepo, tag):
        """Protect an image repo TAG against deletion"""
        return self._protect(self.reposdir + "/" + imagerepo + "/" + tag)

    def unprotect_imagerepo(self, imagerepo, tag):
        """Removes the deletion protection"""
        return self._unprotect(self.reposdir + "/" + imagerepo + "/" + tag)

    def isprotected_imagerepo(self, imagerepo, tag):
        """See if this image TAG is protected against deletion"""
        return self._isprotected(self.reposdir + "/" + imagerepo + "/" + tag)

    def cd_imagerepo(self, imagerepo, tag):
        """Select an image TAG for further operations"""
        if imagerepo and tag:
            tag_dir = self.reposdir + "/" + imagerepo + "/" + tag
            if os.path.exists(tag_dir):
                if self._is_tag(tag_dir):
                    self.cur_repodir = self.reposdir + "/" + imagerepo
                    self.cur_tagdir = self.cur_repodir + "/" + tag
                    return self.cur_tagdir
        return ""

    def _find(self, filename, in_dir):
        """is a specific layer filename referenced by another image TAG"""
        found_list = []
        if FileUtil(in_dir).isdir():
            for fullname in os.listdir(in_dir):
                f_path = in_dir + '/' + fullname
                if os.path.islink(f_path):
                    if filename in fullname:       # match .layer or .json
                        found_list.append(f_path)  # found reference to layer
                elif os.path.isdir(f_path):
                    found_list.extend(self._find(filename, f_path))
        return found_list

    def _inrepository(self, filename):
        """Check if a given file is in the repository"""
        return self._find(filename, self.reposdir)

    def _remove_layers(self, tag_dir, force):
        """Remove link to image layer and corresponding layer
        if not being used by other images
        """
        for fname in os.listdir(tag_dir):
            f_path = tag_dir + '/' + fname  # link to layer
            if os.path.islink(f_path):
                linkname = os.readlink(f_path)
                layer_file = tag_dir + '/' + linkname
                if not FileUtil(f_path).remove() and not force:
                    return False
                if not self._inrepository(os.path.basename(linkname)):
                    # removing actual layers not reference by other repos
                    if not FileUtil(layer_file).remove() and not force:
                        return False
        return True

    def del_imagerepo(self, imagerepo, tag, force=False):
        """Delete an image repository and its layers"""
        tag_dir = self.cd_imagerepo(imagerepo, tag)
        if (tag_dir and
                self._remove_layers(tag_dir, force) and
                FileUtil(tag_dir).remove(recursive=True)):
            self.cur_repodir = ""
            self.cur_tagdir = ""
            while imagerepo:
                FileUtil(self.reposdir + '/' + imagerepo).rmdir()
                imagerepo = "/".join(imagerepo.split("/")[:-1])
            return True
        return False

    def _get_tags(self, tag_dir):
        """Get image tags from repository
        The tags identify actual usable containers
        """
        tag_list = []
        if FileUtil(tag_dir).isdir():
            for fname in os.listdir(tag_dir):
                f_path = tag_dir + '/' + fname
                if self._is_tag(f_path):
                    tag_list.append(
                        (tag_dir.replace(self.reposdir + '/', ""), fname))
                elif os.path.isdir(f_path):
                    tag_list.extend(self._get_tags(f_path))
        return tag_list

    def get_imagerepos(self):
        """get all images repositories with tags"""
        return self._get_tags(self.reposdir)

    def get_layers(self, imagerepo, tag):
        """Get all layers for a given image image tag"""
        layers_list = []
        tag_dir = self.cd_imagerepo(imagerepo, tag)
        if tag_dir:
            for fname in os.listdir(tag_dir):
                filename = tag_dir + "/" + fname
                if os.path.islink(filename):
                    size = FileUtil(filename).size()
                    layers_list.append((filename, size))
        return layers_list

    def add_image_layer(self, filename, linkname=None):
        """Add a layer to an image TAG"""
        if not self.cur_tagdir:
            return False
        if not os.path.exists(filename):
            return False
        if not os.path.exists(self.cur_tagdir):
            return False
        if linkname:
            linkname = self.cur_tagdir + '/' + os.path.basename(linkname)
        else:
            linkname = self.cur_tagdir + '/' + os.path.basename(filename)
        if os.path.islink(linkname):
            FileUtil(linkname).remove()
        self._symlink(filename, linkname)
        return True

    def setup_imagerepo(self, imagerepo):
        """Create directory for an image repository"""
        if not imagerepo:
            return None
        directory = self.reposdir + "/" + imagerepo
        try:
            if not os.path.exists(directory):
                os.makedirs(directory)
                self.cur_repodir = directory
                return True

            self.cur_repodir = directory
            return False
        except (IOError, OSError):
            return None

    def setup_tag(self, tag):
        """Create directory structure for an image TAG
        to be invoked after setup_imagerepo()
        """
        directory = self.cur_repodir + "/" + tag
        try:
            if not os.path.exists(directory):
                os.makedirs(directory)
            self.cur_tagdir = directory
            out_tag = open(directory + "/TAG", 'w', encoding='utf-8')
        except (IOError, OSError):
            return False
        else:
            out_tag.write(self.cur_repodir + ":" + tag)
            out_tag.close()
        return True

    def set_version(self, version):
        """Set the version of the image TAG repository currently
        it supports Docker images with versions v1 and v2
        to be invoked after setup_tag()
        """
        if not (self.cur_repodir and self.cur_tagdir):
            return False
        if not os.path.exists(self.cur_repodir):
            return False
        if not os.path.exists(self.cur_tagdir):
            return False
        directory = self.cur_tagdir
        if (os.path.exists(directory + "/v1") and version != "v1" or
                os.path.exists(directory + "/v2") and version != "v2"):
            if len(os.listdir(directory)) == 1:
                try:
                    FileUtil(directory + "/v1").remove()
                    FileUtil(directory + "/v2").remove()
                except (IOError, OSError):
                    pass
                if os.listdir(directory):
                    return False
        try:
            # Create version file
            open(directory + "/" + version, 'a', encoding='utf-8').close()
        except (IOError, OSError):
            return False
        return True

    def _get_image_attributes_v1(self, directory):
        """Get image attributes from image directory in v1 format"""
        files = []
        layer_list = self.load_json("ancestry")
        if layer_list:
            for layer_id in reversed(layer_list):
                layer_file = directory + '/' + layer_id + ".layer"
                if not os.path.exists(layer_file):
                    return (None, None)
                files.append(layer_file)
            json_file_list = [directory + "/container.json",
                              directory + '/' + layer_list[0] + ".json", ]
            for json_file in json_file_list:
                if os.path.exists(json_file):
                    container_json = self.load_json(json_file)
                    return (container_json, files)
        return (None, None)

    def _get_image_attributes_v2_s1(self, directory, manifest):
        """Get image attributes from image directory in v2 schema 1 format"""
        files = []
        for layer in reversed(manifest["fsLayers"]):
            layer_file = directory + '/' + layer["blobSum"]
            if not os.path.exists(layer_file):
                return (None, None)
            files.append(layer_file)
        try:
            json_string = manifest["history"][0]["v1Compatibility"].strip()
            container_json = json.loads(json_string)
        except (IOError, OSError, AttributeError, ValueError, TypeError,
                IndexError, KeyError):
            return (None, files)
        return (container_json, files)

    def _get_image_attributes_v2_s2(self, directory, manifest):
        """Get image attributes from image directory in v2 schema 2 format"""
        files = []
        for layer in manifest["layers"]:
            layer_file = directory + '/' + layer["digest"]
            if not os.path.exists(layer_file):
                return (None, None)
            files.append(layer_file)
        try:
            json_file = directory + '/' + manifest["config"]["digest"]
            container_json = json.loads(FileUtil(json_file).getdata('r'))
        except (IOError, OSError, AttributeError, ValueError, TypeError,
                IndexError, KeyError):
            return (None, files)
        return (container_json, files)

    def get_image_attributes(self):
        """Load attributes from image TAGs that have been previously
        selected via cd_imagerepo(). Supports images of type v1 and v2.
        Returns: (container JSON, list of layer files).
        """
        directory = self.cur_tagdir
        if os.path.exists(directory + "/v1"):   # if dockerhub API v1
            return self._get_image_attributes_v1(directory)

        if os.path.exists(directory + "/v2"):  # if dockerhub API v1
            manifest = self.load_json("manifest")
            if manifest and "fsLayers" in manifest:
                return self._get_image_attributes_v2_s1(directory, manifest)
            if manifest and "layers" in manifest:
                return self._get_image_attributes_v2_s2(directory, manifest)

        return (None, None)

    def save_json(self, filename, data):
        """Save container json to a file in the image TAG directory
        that has been previously selected via cd_imagerepo()
        or if the file starts with "/" to that specific file.
        """
        if filename.startswith("/"):
            out_filename = filename
        else:
            if not (self.cur_repodir and self.cur_tagdir):
                return False
            if not os.path.exists(self.cur_repodir):
                return False
            if not os.path.exists(self.cur_tagdir):
                return False
            out_filename = self.cur_tagdir + "/" + filename
        outfile = None
        try:
            outfile = open(out_filename, 'w', encoding='utf-8')
            json.dump(data, outfile)
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            if outfile:
                outfile.close()
            return False
        outfile.close()
        return True

    def load_json(self, filename):
        """Load container json from a file in the image TAG directory
        that has been previously selected via cd_imagerepo()
        or if the file starts with '/' from that specific file.
        """
        if filename.startswith('/'):
            in_filename = filename
        else:
            if not (self.cur_repodir and self.cur_tagdir):
                return False
            if not os.path.exists(self.cur_repodir):
                return False
            if not os.path.exists(self.cur_tagdir):
                return False
            in_filename = self.cur_tagdir + '/' + filename
        json_obj = None
        infile = None
        try:
            infile = open(in_filename, 'r', encoding='utf-8')
            json_obj = json.load(infile)
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            pass
        if infile:
            infile.close()
        return json_obj

    def _load_structure(self, imagetagdir):
        """Scan the repository structure of a given image tag"""
        structure = {}
        structure["repolayers"] = {}
        if FileUtil(imagetagdir).isdir():
            for fname in os.listdir(imagetagdir):
                f_path = imagetagdir + '/' + fname
                if fname == "ancestry":
                    structure["ancestry"] = self.load_json(f_path)

                if fname == "manifest":
                    structure["manifest"] = self.load_json(f_path)

                if len(fname) >= 64:
                    layer_id = fname.replace(".json", "").replace(".layer", "")
                    if layer_id not in structure["repolayers"]:
                        structure["repolayers"][layer_id] = {}
                    if fname.endswith("json"):
                        structure["repolayers"][layer_id]["json"] = \
                            self.load_json(f_path)
                        structure["repolayers"][layer_id]["json_f"] = f_path
                        structure["has_json_f"] = True
                    elif fname.endswith("layer"):
                        structure["repolayers"][layer_id]["layer_f"] = f_path
                    elif ':' in fname:
                        structure["repolayers"][layer_id]["layer_f"] = f_path
                    else:
                        Msg().out("Warning: unkwnon file in layer:", f_path,
                                  l=Msg.WAR)
                elif fname in ("TAG", "v1", "v2", "PROTECT", "container.json"):
                    pass

                else:
                    Msg().out("Warning: unkwnon file in image:", f_path, l=Msg.WAR)

        return structure

    def _find_top_layer_id(self, structure, my_layer_id=""):
        """Find the id of the top layer of a given image tag in a
        structure produced by _load_structure()
        """
        if "repolayers" not in structure:
            return ""

        if not my_layer_id:
            if sys.version_info[0] >= 3:
                my_layer_id = list(structure["repolayers"].keys())[0]
            else:
                my_layer_id = structure["repolayers"].keys()[0]

        found = ""
        for layer_id in structure["repolayers"]:
            if "json" not in structure["repolayers"][layer_id]:   # v2
                continue
            if "parent" not in structure["repolayers"][layer_id]["json"]:
                continue
            if (my_layer_id ==
                    structure["repolayers"][layer_id]["json"]["parent"]):
                found = self._find_top_layer_id(structure, layer_id)
                break

        if not found:
            return my_layer_id

        return found

    def _sorted_layers(self, structure, top_layer_id):
        """Return the image layers sorted"""
        sorted_layers = []
        next_layer = top_layer_id
        while next_layer:
            sorted_layers.append(next_layer)
            if "json" not in structure["repolayers"][next_layer]:   # v2
                break

            if "parent" not in structure["repolayers"][next_layer]["json"]:
                break

            next_layer = structure["repolayers"][next_layer]["json"]["parent"]
            if not next_layer:
                break

        return sorted_layers

    def _split_layer_id(self, layer_id):
        """Split layer_id (sha256:xxxxx)"""
        if ':' in layer_id:
            return layer_id.split(":", 1)

        return ("", layer_id)

    def _verify_layer_file(self, structure, layer_id):
        """Verify layer file in repository"""
        (layer_algorithm, layer_hash) = self._split_layer_id(layer_id)
        layer_f = structure["repolayers"][layer_id]["layer_f"]
        if not (os.path.exists(layer_f) and
                os.path.islink(layer_f)):
            Msg().err("Error: layer data file symbolic link not found",
                      layer_id)
            return False
        if not os.path.exists(self.cur_tagdir + '/' +
                              os.readlink(layer_f)):
            Msg().err("Error: layer data file not found")
            return False
        if "gzip" in OSInfo('/').get_filetype(layer_f):
            if not FileUtil(layer_f).verify_tar():
                Msg().err("Error: layer tar verify failed:", layer_f)
                return False
        if layer_algorithm:
            layer_f_chksum = ChkSUM().hash(layer_f, layer_algorithm)
            if layer_f_chksum and layer_f_chksum != layer_hash:
                Msg().err("Error: layer file chksum failed:", layer_f)
                return False
        return True

    def _verify_image_v1(self, structure):
        """Verify the structure of a v1 image repository"""
        Msg().out("Info: finding top layer id", l=Msg.INF)
        top_layer_id = self._find_top_layer_id(structure)
        if not top_layer_id:
            Msg().err("Error: finding top layer id")
            return False
        layers_list = self._sorted_layers(structure, top_layer_id)
        layer = iter(layers_list)
        status = True
        for ancestry_layer in structure["ancestry"]:
            if sys.version_info[0] >= 3:
                verify_layer = next(layer)
            else:
                verify_layer = layer.next()
            if ancestry_layer != verify_layer:
                Msg().err("Error: ancestry and layers do not match",
                          ancestry_layer, verify_layer)
                status = False
                continue
        return status

    def _verify_image_v2_s1(self, structure):
        """Verify the structure of a v2 schema 1 image repository"""
        status = True
        for manifest_layer in structure["manifest"]["fsLayers"]:
            if manifest_layer["blobSum"] not in structure["repolayers"]:
                Msg().err("Error: layer in manifest does not exist",
                          "in repo", manifest_layer["blobSum"])
                status = False
                continue
        return status

    def _verify_image_v2_s2(self, structure):
        """Verify the structure of a v2 schema 2 image repository"""
        status = True
        for manifest_layer in structure["manifest"]["layers"]:
            if manifest_layer["digest"] not in structure["repolayers"]:
                Msg().err("Error: layer in manifest does not exist",
                          "in repo", manifest_layer["blobSum"])
                status = False
                continue
        return status

    def verify_image(self):
        """Verify the structure of an image repository"""
        Msg().out("Info: loading structure", l=Msg.INF)
        structure = self._load_structure(self.cur_tagdir)
        if not structure:
            Msg().err("Error: load of image tag structure failed")
            return False
        Msg().out("Info: verifying layers", l=Msg.INF)
        status = True
        if "ancestry" in structure and "has_json_f" in structure:
            status = self._verify_image_v1(structure)
        elif "manifest" in structure:
            if "fsLayers" in structure["manifest"]:
                status = self._verify_image_v2_s1(structure)
            elif "layers" in structure["manifest"]:
                status = self._verify_image_v2_s2(structure)
        for layer_id in structure["repolayers"]:
            if "layer_f" not in structure["repolayers"][layer_id]:
                Msg().err("Error: layer file not found in structure",
                          layer_id)
                status = False
                continue
            layer_status = self._verify_layer_file(structure, layer_id)
            if not layer_status:
                status = False
                continue
            Msg().out("Info: layer ok:", layer_id, l=Msg.INF)
        return status
